#!/usr/bin/env python
import threading
from functools import wraps
import inspect
import warnings


__author__ = 'Darryl Oatridge'


def singleton(make_instance):
    """Pattern to support a threadsafe singleton.

    Usage: in your singleton override the parent __new__(cls) with the following
        @singleton
        def __new__(cls):
            return super().__new__(cls)

    """
    __lock = threading.Lock()
    __instance = None

    @wraps(make_instance)
    def __new__(cls, *args, **kwargs):
        nonlocal __instance

        if __instance is None:
            with __lock:
                if __instance is None:
                    __instance = make_instance(cls, *args, **kwargs)

        return __instance

    return __new__


def deprecated(reason):
    """
    Much thanks to Laurent LAPORTE for this code from Stack Overflow
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """
    string_types = (type(b''), type(u''))

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "The class {name} has been deprecated. ({reason})."
            else:
                fmt1 = "The function {name} has been deprecated.({reason})."

            @wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter('always', DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2
                )
                warnings.simplefilter('default', DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "The class {name} has been depricated."
        else:
            fmt2 = "The function {name} has been depricated."

        @wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2
            )
            warnings.simplefilter('default', DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))



import unittest
import os
import shutil

from ds_core.handlers.abstract_handlers import ConnectorContract
from ds_core.properties.abstract_properties import AbstractPropertyManager
from ds_core.properties.property_manager import PropertyManager

from test.intent.pyarrow_intent_model import PyarrowIntentModel


class ControlPropertyManager(AbstractPropertyManager):

    def __init__(self, task_name: str, creator: str=None):
        # set additional keys
        root_keys = []
        knowledge_keys = []
        creator = creator if isinstance(creator, str) else 'default'
        super().__init__(task_name=task_name, root_keys=root_keys, knowledge_keys=knowledge_keys, creator=creator)

    @classmethod
    def manager_name(cls) -> str:
        return str(cls.__name__).lower().replace('propertymanager', '')


class IntentModelTest(unittest.TestCase):

    def setUp(self):
        self.connector = ConnectorContract(uri='contracts/config_contract.pq?sep=.&encoding=Latin1',
                                           module_name='ds_core.handlers.pyarrow_handlers',
                                           handler='PyarrowPersistHandler')
        try:
            os.makedirs('contracts')
        except:
            pass
        PropertyManager._remove_all()
        self.pm = ControlPropertyManager('test_abstract_properties')
        self.pm.set_property_connector(self.connector)

    def tearDown(self):
        try:
            shutil.rmtree('contracts')
        except:
            pass

    def test_runs(self):
        """Basic smoke test"""
        PyarrowIntentModel(property_manager=self.pm)

    def test_run_intent_pipeline(self):
        model = PyarrowIntentModel(property_manager=self.pm)
        canonical = {'A': [1,4,7], 'B': [4,5,9]}
        result = model.run_intent_pipeline(canonical=canonical, inplace=False)
        self.assertDictEqual(canonical, result)
        model.to_str_type(data=canonical, headers=['B'])
        result = model.run_intent_pipeline(canonical=canonical)
        self.assertDictEqual({'a': [1, 4, 7]}, result)

    def test_run_intent_pipeline_levels(self):
        model = PyarrowIntentModel(property_manager=self.pm)
        data = {'A': [1,4,7], 'B': [4,5,9]}
        model.to_remove(data=data, headers=['B'], inplace=False, intent_level=0)
        model.auto_clean_header(data=data, case='lower', inplace=False, intent_level=1)
        result = model.run_intent_pipeline(canonical=data, inplace=False, intent_levels=0)
        self.assertDictEqual({'A': [1, 4, 7]}, result)
        result = model.run_intent_pipeline(canonical=data, inplace=False, intent_levels=1)
        self.assertDictEqual({'a': [1,4,7], 'b': [4,5,9]}, result)
        result = model.run_intent_pipeline(canonical=data, inplace=False, intent_levels=[0, 1, 2])
        self.assertDictEqual({'a': [1,4,7]}, result)




if __name__ == '__main__':
    unittest.main()



import os
import unittest
from datetime import datetime
from pprint import pprint

from ds_core.handlers.abstract_handlers import ConnectorContract


class ConnectorContractTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_runs(self):
        """Basic smoke test"""
        ConnectorContract('uri', 'module_name', 'handler', kwarg1='One')

    def test_connector_environ(self):
        os.environ['HADRON_TESTER'] = 'work/tester'
        os.environ['MODULE_ROOT'] = 'tests'
        os.environ['HANDLER'] = 'PyarrowPersistHandler'
        os.environ['VERSION'] = '1.00'
        os.environ['ENCODING'] = 'Latin1'
        os.environ['USERNAME'] = 'dtester'
        os.environ['PASSWORD'] = 'pa22w0rd'
        os.environ['BUCKET'] = 'aistac'
        sc = ConnectorContract('s3://${USERNAME}:${PASSWORD}@${BUCKET}:8080/${HADRON_TESTER}/uri?encoding=${ENCODING}&ignore_errors=True',
                               '${MODULE_ROOT}.handlers.python_handlers',
                               '${HANDLER}',version='v${VERSION}', kwarg1='${HADRON_TESTER}')
        # raw
        self.assertEqual(sc.raw_uri, 's3://${USERNAME}:${PASSWORD}@${BUCKET}:8080/${HADRON_TESTER}/uri?encoding=${ENCODING}&ignore_errors=True')
        self.assertEqual(sc.raw_module_name, '${MODULE_ROOT}.handlers.python_handlers')
        self.assertEqual(sc.raw_handler, '${HANDLER}')
        self.assertEqual(sc.raw_version, 'v${VERSION}')
        self.assertDictEqual(sc.raw_kwargs, {'kwarg1': '${HADRON_TESTER}'})
        # parsed
        self.assertEqual(sc.uri, 's3://dtester:pa22w0rd@aistac:8080/work/tester/uri?encoding=Latin1&ignore_errors=True')
        self.assertEqual(sc.module_name, 'tests.handlers.python_handlers')
        self.assertEqual(sc.handler, 'PyarrowPersistHandler')
        self.assertEqual(sc.version, 'v1.00')
        self.assertDictEqual(sc.kwargs, {'kwarg1': 'work/tester'})
        self.assertEqual(sc.address, 's3://dtester:pa22w0rd@aistac:8080/work/tester/uri')
        self.assertEqual(sc.hostname, 'aistac')
        self.assertEqual(sc.netloc, 'dtester:pa22w0rd@aistac:8080')
        self.assertEqual(sc.password, 'pa22w0rd')
        self.assertEqual(sc.path, '/work/tester/uri')
        self.assertEqual(sc.port, 8080)
        self.assertDictEqual(sc.query, {'encoding': 'Latin1', 'ignore_errors': 'True'})
        self.assertEqual(sc.schema, 's3')
        self.assertEqual(sc.username, 'dtester')
        os.environ.pop('HADRON_TESTER')
        os.environ.pop('MODULE_ROOT')
        os.environ.pop('HANDLER')
        os.environ.pop('VERSION')
        os.environ.pop('ENCODING')
        os.environ.pop('USERNAME')
        os.environ.pop('PASSWORD')
        os.environ.pop('BUCKET')

    def test_immutable(self):
        sc = ConnectorContract('uri', 'module_name', 'handler', kwarg1='One')
        control = {'uri': 'uri', 'uri_parsed': {'address': 'uri','schema': '', 'netloc': '', 'path': 'uri', 'params': [], 'query': {}, 'fragment': ''},
                   'module_name': 'module_name',
                   'handler': 'handler',
                   'version': "0.0.1",
                   'kwargs': {'kwarg1': 'One'}}
        self.assertDictEqual(control, sc.to_dict())
        with self.assertRaises(AttributeError) as context:
            sc.raw_uri = "new"
        self.assertTrue("The attribute 'raw_uri' is immutable once set and can not be changed" in str(context.exception))

    def test_empty_uri(self):
        with self.assertRaises(ValueError) as context:
            ConnectorContract(None, None, None)
        self.assertTrue("The uri must be a valid string" in str(context.exception))
        with self.assertRaises(ValueError) as context:
            ConnectorContract("", None, None)
        self.assertTrue("The uri must be a valid string" in str(context.exception))
        sc = ConnectorContract('a', None, None)
        result = sc.to_dict()
        control = {'uri': 'a', 'uri_parsed': {'address': 'a', 'schema': '', 'netloc': '', 'path': 'a', 'params': [],
                                               'query': {}, 'fragment': ''},
                   'module_name': None,
                   'handler': None,
                   'version': "0.0.1",
                   'kwargs': {}}
        self.assertDictEqual(control, result)

    def test_min_uri(self):
        uri = "file.csv"
        sc = ConnectorContract(uri, 'base.module_name', 'Handler')
        result = sc.to_dict()
        control = {'uri': 'file.csv', 'uri_parsed': {'address': 'file.csv', 'schema': '', 'netloc': '',
                                                     'path': 'file.csv', 'params': [], 'query': {}, 'fragment': ''},
                   'module_name': 'base.module_name',
                   'handler': 'Handler',
                   'version': "0.0.1",
                   'kwargs': {}}
        self.assertDictEqual(control, result)

    def test_full_uri(self):
        uri = "http://user:pass@NetLoc:80/path/file.csv;parameters2;param3?query=argument&query2=attr#fragment"
        sc = ConnectorContract(uri, 'base.module_name', 'Handler', kwarg1='One')
        result = sc.to_dict()
        control = {'uri': 'http://user:pass@NetLoc:80/path/file.csv;parameters2;param3?query=argument&query2=attr#fragment',
                   'uri_parsed': {'address': 'http://user:pass@netloc:80/path/file.csv', 'schema': 'http',
                                  'netloc': 'user:pass@NetLoc:80', 'path': '/path/file.csv',
                                  'params': ['parameters2', 'param3'], 'query': {'query': 'argument', 'query2': 'attr'},
                                  'fragment': 'fragment', 'username': 'user', 'password': 'pass', 'hostname': 'netloc',
                                  'port': 80},
                   'module_name': 'base.module_name',
                   'handler': 'Handler',
                   'version': "0.0.1",
                   'kwargs': {'kwarg1': 'One'}}
        self.assertDictEqual(control, result)

    def test_get(self):
        uri = "http://user:pass@NetLoc:80/path/file.csv;parameters2;param3?query1=value1&query2=value2#fragment"
        cc = ConnectorContract(uri, 'base.module_name', 'Handler', kwarg1='kwOne', kwarg2='kwTwo')
        self.assertEqual('value1', cc.get_key_value('query1'))
        self.assertEqual('kwTwo', cc.get_key_value('kwarg2'))
        self.assertEqual('NoValue', cc.get_key_value('kwarg3', 'NoValue'))
        self.assertEqual({'query1': 'value1', 'query2': 'value2'}, cc.query)
        self.assertEqual({'kwarg1': 'kwOne', 'kwarg2': 'kwTwo'}, cc.kwargs)

    def test_parse_address(self):
        uri = "http://user:pass@NetLoc:80/path/file.csv;parameters2;param3?query1=value1&query2=value2#fragment"
        result = ConnectorContract.parse_address(uri=uri)
        self.assertEqual('http://user:pass@netloc:80/path/file.csv', result)
        result = ConnectorContract.parse_address(uri=uri, with_credentials=True)
        self.assertEqual('http://user:pass@netloc:80/path/file.csv', result)
        result = ConnectorContract.parse_address(uri=uri, with_port=True)
        self.assertEqual('http://user:pass@netloc:80/path/file.csv', result)
        # False
        result = ConnectorContract.parse_address(uri=uri, with_credentials=False)
        self.assertEqual('http://netloc:80/path/file.csv', result)
        result = ConnectorContract.parse_address(uri=uri, with_port=False)
        self.assertEqual('http://user:pass@netloc/path/file.csv', result)
        result = ConnectorContract.parse_address(uri=uri, with_credentials=False, with_port=False)
        self.assertEqual('http://netloc/path/file.csv', result)
        # minimal
        uri = "file.csv"
        result = ConnectorContract.parse_address(uri=uri)
        self.assertEqual('file.csv', result)
        result = ConnectorContract.parse_address(uri=uri, with_credentials=True, with_port=True)
        self.assertEqual('file.csv', result)
        result = ConnectorContract.parse_address(uri=uri, with_credentials=False, with_port=False)
        self.assertEqual('file.csv', result)

    def test_parse_elemets(self):
        uri = "http://user:pass@NetLoc:80/path/file.csv;parameters2;param3?query1=value1&query2=value2#fragment"
        schema, netloc, path = ConnectorContract.parse_address_elements(uri=uri)
        self.assertEqual('http', schema)
        self.assertEqual('user:pass@netloc:80', netloc)
        self.assertEqual('/path/file.csv', path)
        # False
        schema, netloc, path = ConnectorContract.parse_address_elements(uri=uri, with_credentials=False, with_port=False)
        self.assertEqual('http', schema)
        self.assertEqual('netloc', netloc)
        self.assertEqual('/path/file.csv', path)

    def test_parse_query(self):
        uri = "file.csv"
        query = ConnectorContract.parse_query(uri)
        self.assertEqual({}, query)
        uri = "http://user:pass@NetLoc:80/path/file.csv;parameters2;param3?query1=value1&query2=value2#fragment"
        query = ConnectorContract.parse_query(uri)
        self.assertEqual({'query1': 'value1', 'query2': 'value2'}, query)

    def test_uri(self):
        uri = "s3://bucket/path/aistac_manager_task_connector.parquet"
        cc = ConnectorContract(uri, 'module.name', 'Handler')
        self.assertEqual(uri, cc.uri)
        # VERSION
        uri = "s3://bucket/path/aistac_manager_task_connector${VERSION}.parquet"
        cc = ConnectorContract(uri, 'module.name', 'Handler')
        self.assertEqual("s3://bucket/path/aistac_manager_task_connector_0.0.1.parquet", cc.uri)
        cc = ConnectorContract(uri, 'module.name', 'Handler', version="v1.01")
        self.assertEqual("s3://bucket/path/aistac_manager_task_connector_v1.01.parquet", cc.uri)
        # DATETIME
        uri = "s3://bucket/path/aistac_manager_task_connector${TO_DAYS}.parquet"
        cc = ConnectorContract(uri, 'module.name', 'Handler')
        control = f"s3://bucket/path/aistac_manager_task_connector{datetime.now().strftime('_%Y%m%d')}.parquet"
        self.assertEqual(control, cc.uri)
        # VERSION & DATETIME
        uri = "s3://bucket/path/aistac_manager_task_connector${VERSION}${TO_DAYS}.parquet"
        cc = ConnectorContract(uri, 'module.name', 'Handler', version="v1.01")
        control = f"s3://bucket/path/aistac_manager_task_connector_v1.01{datetime.now().strftime('_%Y%m%d')}.parquet"
        self.assertEqual(control, cc.uri)

    def test_encode_query(self):
        result = ConnectorContract.encode_query(name='fred blogs', age=35)
        self.assertEqual('name=fred+blogs&age=35', result)
        result = ConnectorContract.encode_query()
        self.assertEqual('', result)

    def test_version(self):
        uri = "s3://bucket/path/aistac_manager_task_connector.parquet"
        cc = ConnectorContract(uri, 'module.name', 'Handler')
        self.assertEqual("0.0.1", cc.version)
        cc = ConnectorContract(uri, 'module.name', 'Handler', version='v1.00')
        self.assertEqual("v1.00", cc.version)


if __name__ == '__main__':
    unittest.main()



from ds_core.handlers.abstract_handlers import HandlerFactory, ConnectorContract
import unittest


class HandlerFactoryTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_runs(self):
        """Basic smoke test"""
        HandlerFactory()

    def test_exceptions(self):
        # Bad module
        connector_contract = ConnectorContract(uri='example.csv;file_type=csv',
                                               module_name='ds_core.handlers.none',
                                               handler='PyarrowSourceHandler',
                                               kwargs={'sep': ',', 'encoding': 'latin1'})
        with self.assertRaises(ModuleNotFoundError) as context:
            HandlerFactory.instantiate(connector_contract)
        self.assertTrue("The module 'ds_core.handlers.none' could not be found" in str(context.exception))
        # bad handler
        connector_contract = ConnectorContract(uri='example.csv', module_name='ds_core.handlers.pyarrow_handlers',
                                               handler='none')
        with self.assertRaises(ImportError) as context:
            HandlerFactory.instantiate(connector_contract)
        self.assertTrue("The handler 'none' could not be found in the module 'ds_core.handlers.pyarrow_handlers'" in str(context.exception))
        # Everything correct
        connector_contract = ConnectorContract(uri='example.csv', module_name='ds_core.handlers.pyarrow_handlers',
                                               handler='PyarrowSourceHandler')
        handler = HandlerFactory.instantiate(connector_contract)
        self.assertEqual("<class 'ds_core.handlers.pyarrow_handlers.PyarrowSourceHandler'>", str(type(handler)))

    def test_handler_check(self):
        # Module
        result = HandlerFactory.check_module('ds_core.handlers.pyarrow_handlers')
        self.assertTrue(result)
        result = HandlerFactory.check_module('ds_core.handlers.none')
        self.assertFalse(result)
        # Handdler
        result = HandlerFactory.check_handler(module_name='ds_core.handlers.pyarrow_handlers', handler='PyarrowSourceHandler')
        self.assertTrue(result)
        result = HandlerFactory.check_handler(module_name='ds_core.handlers.None', handler='PyarrowSourceHandler')
        self.assertFalse(result)
        result = HandlerFactory.check_handler(module_name='ds_core.handlers.pyarrow_handlers', handler='None')
        self.assertFalse(result)


if __name__ == '__main__':
    unittest.main()



import os
import shutil
import unittest
from pprint import pprint

from ds_core.components.core_commons import CoreCommons
from ds_core.handlers.abstract_handlers import ConnectorContract

from ds_core.properties.abstract_properties import AbstractPropertyManager, AbstractProperty
from ds_core.properties.property_manager import PropertyManager


class ControlPropertyManager(AbstractPropertyManager):

    def __init__(self, task_name: str, root_keys: list=None, knowledge_keys: list=None, creator: str=None):
        root_keys = root_keys if isinstance(root_keys, list) else []
        knowledge_keys = knowledge_keys if isinstance(knowledge_keys, list) else []
        creator = creator if isinstance(creator, str) else 'default'
        root_keys += ['cleaners']
        super().__init__(task_name, root_keys, knowledge_keys, creator)

    @classmethod
    def manager_name(cls) -> str:
        return str(cls.__name__).lower().replace('propertymanager', '')


class AbstractPropertiesManagerTest(unittest.TestCase):
    """Test: """

    def setUp(self):
        self.connector = ConnectorContract(uri='works/config_contract.parquet?sep=.&encoding=Latin1',
                                           module_name='ds_core.handlers.pyarrow_handlers',
                                           handler='PyarrowPersistHandler', name='darryl', password='mypass')
        try:
            os.makedirs('works')
        except:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('works')
        except:
            pass

    def test_runs(self):
        """Basic smoke test"""
        ControlPropertyManager('test_abstract_properties')

    def test_manager_name(self):
        manager = ControlPropertyManager('test_abstract_properties')
        self.assertEqual('control', manager.manager_name())

    def test_creator(self):
        manager = ControlPropertyManager('test_abstract_properties')
        self.assertEqual('default', manager.creator)
        manager = ControlPropertyManager('test_abstract_properties', creator='fbloggs')
        self.assertEqual('fbloggs', manager.creator)

    def test_keys(self):
        augmented_keys = ['observations', 'notes']
        manager = ControlPropertyManager('test_abstract_properties', knowledge_keys=augmented_keys)
        result = manager.KEY.keys()
        self.assertIn('knowledge.observations_key', result)
        self.assertIn('knowledge.notes_key', result)
        self.assertIn("knowledge.schema_key", result)
        self.assertIn('cleaners_key', result)
        self.assertIn('version_key', result)
        control = 'control.test_abstract_properties.version'
        result = manager.KEY.version_key
        self.assertEqual(control, result)
        control = 'control.test_abstract_properties.knowledge.notes'
        result = manager.KEY.knowledge.notes_key
        self.assertEqual(control, result)

    def test_persist(self):
        pm = ControlPropertyManager('test_abstract_properties')
        pm.set(pm.join(pm.KEY.contract_key, 'test_key'), 'test value')
        result = pm.get(pm.join(pm.KEY.contract_key, 'test_key'))
        self.assertEqual('test value', result)
        # create the connection
        pm.set_property_connector(self.connector)
        self.assertFalse(os.path.exists('works/config_contract.p'))
        pm.set_version('test_version')
        pm.persist_properties()
        self.assertTrue(os.path.exists('works/config_contract.parquet'))
        self.assertEqual(['pm_control_test_abstract_properties'], pm.connector_contract_list)
        self.assertEqual('test_version', pm.version)
        pm.reset_all()
        self.assertEqual('0.0.1', pm.version)
        self.assertEqual(['pm_control_test_abstract_properties'], pm.connector_contract_list)

    def test_abstract_key(self):
        keys = ['connectors', 'values']
        ab = AbstractProperty(keys, manager='contract', contract='subset')
        self.assertEqual('contract.subset', ab.contract_key)
        self.assertEqual('contract.subset.connectors', ab.connectors_key)

        ab = AbstractProperty(keys, manager='contract')
        self.assertEqual('contract', ab.manager_key)
        self.assertEqual('contract.connectors', ab.connectors_key)

        ab = AbstractProperty(keys)
        self.assertEqual('connectors', ab.connectors_key)
        self.assertEqual('values', ab.values_key)

    def test_abstract_key_complex(self):
        keys = [{'connectors': ['one', 'two']}]
        ab = AbstractProperty(keys, manager='contract', contract='subset')
        self.assertEqual('contract.subset', ab.contract_key)
        self.assertEqual('contract.subset.connectors', ab.connectors_key)
        self.assertEqual('contract.subset.connectors.one', ab.connectors.one_key)

    def test_reset_abstract_properties(self):
        keys = ['connectors', 'intent', 'overview']
        pm = ControlPropertyManager(task_name='contract', root_keys=keys, knowledge_keys=[])
        pm.set(pm.KEY.intent_key, 'some value')
        pm.set_version('v1')
        control = {'description': "", 'status': 'discovery', 'cleaners': {}, 'connectors': {}, 'intent': 'some value', 'overview': {}, 'run_book': {},'snapshot': {}, 'version': 'v1',
                   'meta': {'class': 'ControlPropertyManager', 'module': pm.__module__.split(".")}, 'knowledge': {'intent': {}, 'schema': {}}}
        self.assertEqual(control, pm.get(pm.KEY.contract_key))
        pm.reset_all()
        control = {'description': "", 'status': 'discovery', 'cleaners': {}, 'connectors': {}, 'intent': {}, 'overview': {}, 'run_book': {},'snapshot': {}, 'version': '0.0.1',
                   'meta': {'class': 'ControlPropertyManager', 'module': pm.__module__.split(".")}, 'knowledge': {'intent': {}, 'schema': {}}}
        self.assertEqual(control, pm.get(pm.KEY.contract_key))

    def test_connection_handler(self):
        pm = ControlPropertyManager('test_abstract_properties', root_keys=[], knowledge_keys=[])
        control_connector = self.connector
        pm.set_connector_contract(connector_name='control', connector_contract=self.connector)
        result = pm.get_connector_contract(connector_name='control')
        self.assertEqual(control_connector, result)
        # connections should be empty as not requested a handler
        self.assertEqual([], pm.connector_handler_list)
        self.assertEqual(['control'], pm.connector_contract_list)
        pm.get_connector_handler(connector_name='control')
        self.assertEqual(['control'], pm.connector_handler_list)
        pm.remove_connector_contract(connector_name='control')
        control = {'description': '', 'status': 'discovery', 'cleaners': {}, 'connectors': {},  'intent': {}, 'run_book': {}, 'snapshot': {}, 'version': '0.0.1',
                   'meta': {'class': 'ControlPropertyManager', 'module': pm.__module__.split(".")}, 'knowledge': {'intent': {}, 'schema': {}}}
        self.assertEqual(control, pm.get(pm.KEY.contract_key))

    def test_set_property(self):
        pm = ControlPropertyManager('test_abstract_properties')
        pm.set_property_connector(self.connector)
        result = pm.get_connector_contract(pm.CONNECTOR_PM_CONTRACT)
        self.assertEqual('works/config_contract.parquet?sep=.&encoding=Latin1', result.raw_uri)
        self.assertEqual('ds_core.handlers.pyarrow_handlers', result.raw_module_name)
        self.assertEqual('PyarrowPersistHandler', result.raw_handler)
        self.assertDictEqual({'name': 'darryl', 'password': 'mypass'}, result.raw_kwargs)
        self.assertEqual('0.0.1', result.version)

    def test_reset(self):
        pm = ControlPropertyManager('test_abstract_properties')
        pm.reset_all()
        pm.set_property_connector(self.connector)
        control = pm.get(pm.KEY.connectors_key)
        pm.set_version('latest')
        version = pm.version
        self.assertEqual('latest', pm.version)
        pm.reset_all()
        result = pm.get(pm.KEY.connectors_key)
        self.assertEqual(control, result)
        self.assertNotEqual(version, pm.version)
        self.assertEqual('0.0.1', pm.version)
        # add new properties and check the connector is not removed
        pm.set(pm.KEY.cleaners_key, {'clean_header': {'case': 'title', 'rename': {'surname: last_name'}}})
        pm.set(pm.KEY.version_key, 'v1.00')
        self.assertEqual(['pm_control_test_abstract_properties'], list(pm.get(pm.KEY.connectors_key).keys()))
        self.assertEqual(['clean_header'], list(pm.get(pm.KEY.cleaners_key).keys()))
        self.assertEqual('v1.00', pm.get(pm.KEY.version_key))
        pm.reset_all()
        self.assertEqual(['pm_control_test_abstract_properties'], list(pm.get(pm.KEY.connectors_key).keys()))
        self.assertEqual({}, pm.get(pm.KEY.cleaners_key))
        self.assertEqual('0.0.1', pm.get(pm.KEY.version_key))

    def test_multi_instance(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        self.assertEqual('0.0.1', dpm.get(dpm.KEY.version_key))
        self.assertEqual([], list(dpm.get(dpm.KEY.cleaners_key).keys()))
        dpm.set_connector_contract('raw_source', connector_contract=self.connector)
        dpm.set(dpm.KEY.cleaners_key, {'clean_header': {'case': 'title', 'rename': {'surname: last_name'}}})
        dpm.set(dpm.KEY.version_key, '5.01')
        self.assertEqual('5.01', dpm.get(dpm.KEY.version_key))
        self.assertEqual(['clean_header'], list(dpm.get(dpm.KEY.cleaners_key).keys()))
        #now load a new instance
        dpm2 = ControlPropertyManager('test_abstract_properties')
        self.assertEqual('5.01', dpm2.get(dpm.KEY.version_key))
        self.assertEqual(['clean_header'], list(dpm.get(dpm.KEY.cleaners_key).keys()))

    def test_connector_handler(self):
        connector_name = 'raw_source'
        dpm = ControlPropertyManager('test_abstract_properties')
        dpm.set_connector_contract(connector_name=connector_name, connector_contract=self.connector)
        handler = dpm.get_connector_handler(connector_name)
        self.assertEqual("<class 'ds_core.handlers.pyarrow_handlers.PyarrowPersistHandler'>", str(type(handler)))

    def test_set_connector_contract(self):
        connector_name = 'raw_source'
        dpm = ControlPropertyManager('test_abstract_properties')
        dpm.set_connector_contract(connector_name=connector_name, connector_contract=self.connector)
        result = dpm.get_connector_contract(connector_name)
        self.assertEqual(self.connector.raw_uri, result.raw_uri)
        self.assertEqual(self.connector.module_name, result.module_name)
        self.assertEqual(self.connector.handler, result.handler)
        self.assertEqual(self.connector.get_key_value('sep'), result.get_key_value('sep'))
        self.assertEqual(self.connector.get_key_value('encoding'), result.get_key_value('encoding'))

    def test_set_version(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        self.assertEqual('0.0.1', dpm.get(dpm.KEY.version_key))
        dpm.set(dpm.KEY.version_key, '1.00')
        self.assertEqual('1.00', dpm.get(dpm.KEY.version_key))
        self.assertEqual(dpm.get(dpm.KEY.version_key), dpm.version)

    def test_set_status(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        self.assertEqual('discovery', dpm.get(dpm.KEY.status_key))
        dpm.set(dpm.KEY.status_key, 'complete')
        self.assertEqual('complete', dpm.get(dpm.KEY.status_key))
        self.assertEqual(dpm.get(dpm.KEY.status_key), dpm.status)

    def test_set_description(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        self.assertEqual('', dpm.get(dpm.KEY.description_key))
        dpm.set(dpm.KEY.description_key, 'This Task description')
        self.assertEqual('This Task description', dpm.get(dpm.KEY.description_key))
        self.assertEqual(dpm.get(dpm.KEY.description_key), dpm.description)

    def test_info(self):
        dpm = ControlPropertyManager('test')
        result = dpm.report_task_meta()
        control = {'contract': 'control', 'task': 'test', 'description': '', 'status': 'discovery', 'version': '0.0.1'}
        self.assertEqual(control, result)
        dpm.set_description('This Task description')
        dpm.set_version('v1.01')
        dpm.set_status('stable')
        result = dpm.report_task_meta()
        control = {'contract': 'control', 'task': 'test', 'description': 'This Task description', 'status': 'stable', 'version': 'v1.01'}
        self.assertEqual(control, result)
        dpm.reset_task_meta()
        result = dpm.report_task_meta()
        control = {'contract': 'control', 'task': 'test', 'description': '', 'status': 'discovery', 'version': '0.0.1'}
        self.assertEqual(control, result)

    def test_set_snapshot(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        dpm.set_snapshot('3.00.001')
        control = ['description', 'status', 'cleaners', 'snapshot', 'intent', 'version', 'run_book','connectors', 'knowledge', 'meta']
        self.assertCountEqual(control, list(dpm.get(dpm.KEY.contract_key).keys()))
        result = list(dpm.get(dpm.KEY.snapshot_key).keys())
        control = ['test_abstract_properties_#3_00_001']
        self.assertCountEqual(control, result)
        result = list(dpm.get(dpm.KEY.snapshot_key).get('test_abstract_properties_#3_00_001').keys())
        # check the snapshot hasn't been copied to the snapshot
        control = ['cleaners', 'description', 'version', 'status', 'connectors', 'intent', 'run_book', 'meta', 'knowledge', 'snapshot']
        self.assertCountEqual(control, result)

    def test_recover_snapshot(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        version = '1.00.000'
        dpm.set(dpm.KEY.version_key, version)
        self.assertEqual(version, dpm.version)
        snap_name = dpm.set_snapshot("test")
        control = '1.00.001'
        dpm.set(dpm.KEY.version_key, control)
        self.assertEqual(control, dpm.version)
        dpm.recover_snapshot(snap_name)
        self.assertEqual(version, dpm.version)

    def test_get_snapshot(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        control = []
        result = dpm.snapshots
        self.assertEqual(control, result)
        dpm.set_snapshot('3.00.001')
        dpm.set_snapshot('test')
        control = ['test_abstract_properties_#3_00_001', 'test_abstract_properties_#test']
        result = dpm.snapshots
        self.assertEqual(control, result)
        dpm2 = ControlPropertyManager('new_abstract_properties')
        dpm2.set_snapshot('2018-06-12')
        control = ['new_abstract_properties_#2018-06-12']
        result = dpm2.snapshots
        self.assertEqual(control, result)
        control = ['test_abstract_properties_#3_00_001', 'test_abstract_properties_#test']
        result = dpm.snapshots
        self.assertEqual(control, result)

    def test_get_knowledge(self):
        keys = ['overview', 'notes']
        manager = ControlPropertyManager('test_abstract_properties', knowledge_keys=keys)
        manager.set_knowledge(text="My Note", label='age', catalog='notes')
        manager.set_knowledge(text="general note", label='comment', catalog='notes')
        manager.set_knowledge(text="and another", label='comment', catalog='notes')
        result = manager.get_knowledge(catalog='overview')
        self.assertEqual({}, result)
        result = manager.get_knowledge(catalog='notes')
        self.assertEqual(['age', 'comment'], CoreCommons.list_formatter(result.keys()))
        result = manager.get_knowledge(catalog='notes', as_list=True)
        self.assertEqual(['age', 'comment'], result)
        result = manager.get_knowledge(catalog='notes', label='comment')
        self.assertEqual(['general note', 'and another'], CoreCommons.list_formatter(result.values()))
        result = manager.get_knowledge(catalog='notes', label='comment', as_list=True)
        self.assertEqual(['general note', 'and another'], result)

    def test_set_and_remove_knowledge(self):
        keys = ['overview', 'notes', 'observations', 'attribute', 'dictionary', 'tor']
        manager = ControlPropertyManager('test_abstract_properties', knowledge_keys=keys)
        manager.set_knowledge(text="My Note", label='age', catalog='notes')
        manager.set_knowledge(text="general note", label='comment', catalog='notes')
        result = manager.knowledge_filter(catalog='notes', label='age')
        self.assertEqual(["My Note"], list(result.get('notes').get('age').values()))
        #remove
        result = manager.reset_knowledge()
        self.assertEqual({}, manager.knowledge_filter(catalog='notes'))

    def test_set_knowledge_with_replace(self):
        keys = ['overview', 'notes', 'observations', 'attribute', 'dictionary', 'tor']
        manager = ControlPropertyManager('test_abstract_properties', knowledge_keys=keys)
        manager.set_knowledge(text="My Note", label='age', catalog='notes')
        result = manager.knowledge_filter(catalog='notes', label='age')
        self.assertEqual(["My Note"], list(result.get('notes').get('age').values()))
        manager.set_knowledge(text="New Note", label='age', catalog='notes', replace=True)
        result = manager.knowledge_filter(catalog='notes', label='age')
        self.assertEqual(["New Note"], list(result.get('notes').get('age').values()))

    def test_has_knowledge(self):
        keys = ['overview', 'notes', 'observations', 'attribute', 'dictionary', 'tor']
        manager = ControlPropertyManager('test_abstract_properties', knowledge_keys=keys)
        result = manager.has_knowledge()
        self.assertFalse(result)
        manager.set_knowledge(catalog='overview', label='label', text='text')
        result = manager.has_knowledge()
        self.assertTrue(result)
        result = manager.has_knowledge(catalog='notes')
        self.assertFalse(result)
        result = manager.has_knowledge(catalog='overview')
        self.assertTrue(result)
        result = manager.has_knowledge(catalog='overview', label='bob')
        self.assertFalse(result)
        result = manager.has_knowledge(catalog='overview', label='label')
        self.assertTrue(result)
        result = manager.has_knowledge(label='label')
        self.assertTrue(result)
        result = manager.has_knowledge(label='bob')
        self.assertFalse(result)

    def test_bulk_upload_and_filter(self):
        keys = ['attributes']
        manager = ControlPropertyManager('test_abstract_properties', knowledge_keys=keys)
        upload = {'Attribute': ['APP_CD', 'APP_DESC', 'BILT_CUST_NBR', 'BU_ID', 'FISC_QTR_ID', 'FISC_WK_ID'],
                  'description': ['Esupport Application Code', 'Esupport Application Description',
                                  'Bill To Customer number', 'Business Unit ID (number that represents the country)',
                                  'Fiscal Quarter', 'Fiscal WK', ]}
        manager.bulk_upload_knowledge(upload, catalog='attributes', label_key='Attribute', text_key='description')
        result = manager.report_notes(catalog='attributes')
        self.assertEqual(['attributes']*6, result.get('section'))
        self.assertCountEqual(upload.get('Attribute'), result.get('label'))
        self.assertCountEqual(upload.get('description'), result.get('text'))

    def test_set_intent(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        with self.assertRaises(ValueError) as context:
            dpm.set_intent(intent_param=[])
        self.assertTrue("The intent section must be a dictionary" in str(context.exception))
        dpm.set_intent(intent_param={})
        control = {}
        self.assertEqual(control, dpm.get_intent())
        # add values
        cleaner1 = {'clean_01': {'headers': 'h1', 'dtype': ['number', 'str']}}
        dpm.set_intent(cleaner1)
        control = {'A': {'0': {'clean_01': {'dtype': ['number', 'str'], 'headers': 'h1'}}}}
        self.assertEqual(control, dpm.get_intent())
        # adding the same cleaner should not change anything
        dpm.set_intent(cleaner1)
        self.assertEqual(control, dpm.get_intent())
        # same cleaner different intent
        cleaner1 = {'clean_01': {'headers': 'h4'}}
        control = {'A': {'0': {'clean_01': {'dtype': ['number', 'str'], 'headers': 'h1'}}}}
        dpm.set_intent(cleaner1, order=0, replace_intent=False)
        self.assertEqual(control, dpm.get_intent())
        control = {'A': {'0': {'clean_01': {'headers': 'h4'}}}}
        dpm.set_intent(cleaner1, order=0, replace_intent=True)
        self.assertEqual(control, dpm.get_intent())
        # second cleaner
        cleaner2 = {'clean_02': {'headers': ['h2', 'h3'], 'regex': 'text'}}
        dpm.set_intent(cleaner2, order=0)
        control = {'A': {'0': {'clean_01': {'headers': 'h4'},
                         'clean_02': {'headers': ['h2', 'h3'], 'regex': 'text'}}}}
        self.assertEqual(control, dpm.get_intent())
        cleaner1 = {'clean_01': {'headers': 'h1', 'dtype': ['number', 'str']}}
        dpm.set_intent(cleaner1, order=0, replace_intent=True)
        control = {'A': {'0': {'clean_01': {'dtype': ['number', 'str'], 'headers': 'h1'},
                         'clean_02': {'headers': ['h2', 'h3'], 'regex': 'text'}}}}
        self.assertEqual(control, dpm.get_intent())

    def test_set_intent_order(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        cleaner1 = {'clean_01': {'headers': 'h1'}}
        dpm.set_intent(cleaner1, order=0)
        control = {'A': {'0': {'clean_01': {'headers': 'h1'}}}}
        self.assertEqual(control, dpm.get_intent())
        dpm.set_intent(cleaner1, order=1)
        control = {'A': {'1': {'clean_01': {'headers': 'h1'}}}}
        self.assertEqual(control, dpm.get_intent())
        # change params
        cleaner1 = {'clean_01': {'headers': 'h3'}}
        dpm.set_intent(cleaner1, order=0)
        control = {'A': {'0': {'clean_01': {'headers': 'h3'}},
                         '1': {'clean_01': {'headers': 'h1'}}}}
        self.assertEqual(control, dpm.get_intent())
        # used -1 same intent, different params
        cleaner1 = {'clean_01': {'headers': 'h4'}}
        dpm.set_intent(cleaner1, order=-1)
        control = {'A': {'0': {'clean_01': {'headers': 'h3'}},
                         '1': {'clean_01': {'headers': 'h1'}},
                         '2': {'clean_01': {'headers': 'h4'}}}}
        self.assertEqual(control, dpm.get_intent())
        # replace just one of them using -1
        cleaner1 = {'clean_01': {'headers': 'h1'}}
        dpm.set_intent(cleaner1, order=-1)
        self.assertEqual(control, dpm.get_intent())

    def test_remove_intent(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        # setup
        dpm.set_intent({'clean_01': {'headers': 'h1'}})
        # remove by label
        self.assertTrue(dpm.has_intent(level='A'))
        dpm.remove_intent('clean_01')
        self.assertEqual({}, dpm.get_intent())
        dpm.set_intent({'clean_01': {'headers': 'h1'}}, order=0)
        dpm.set_intent({'clean_01': {'headers': 'h4'}}, order=1)
        self.assertTrue(dpm.has_intent(level='A'))
        dpm.remove_intent('clean_01')
        self.assertEqual({}, dpm.get_intent())
        # remove by cleaner
        dpm.set_intent({'clean_01': {'headers': 'h1'}}, order=0)
        dpm.set_intent({'clean_01': {'headers': 'h4'}}, order=1)
        self.assertEqual(['0', '1'], CoreCommons.list_formatter(dpm.get_intent(level='A').keys()))
        dpm.remove_intent({'clean_01': {'headers': 'h4'}})
        self.assertTrue(dpm.has_intent(level='A'))
        self.assertEqual(['0'], CoreCommons.list_formatter(dpm.get_intent(level='A').keys()))
        # remove all
        dpm.set_intent({'clean_01': {'headers': 'h1'}})
        dpm.set_intent({'clean_02': {'headers': 'h4'}})
        dpm.set_intent({'clean_03': {'headers': 'h3'}}, level='B')
        dpm.reset_intents()
        self.assertEqual({}, dpm.get_intent())
        # test restricted remove
        dpm.set_intent({'clean_01': {'headers': 'h1'}}, level='A')
        dpm.set_intent({'clean_02': {'headers': 'h4'}}, level='B')
        dpm.remove_intent(level='C')
        dpm.remove_intent(level='A')
        self.assertEqual(['B'], CoreCommons.list_formatter(dpm.get_intent().keys()))
        dpm.remove_intent(level='B')
        self.assertEqual({}, dpm.get_intent())
        # remove all intent in a level
        dpm.set_intent({'clean_01': {'headers': 'h1'}}, level='A', order=-1)
        dpm.set_intent({'clean_01': {'headers': 'h2'}}, level='A', order=-1)
        dpm.set_intent({'clean_01': {'headers': 'h4'}}, level='B')
        self.assertEqual(['0', '1'], CoreCommons.list_formatter(dpm.get_intent(level='A').keys()))
        self.assertEqual(['0'], CoreCommons.list_formatter(dpm.get_intent(level='B').keys()))
        dpm.remove_intent(intent_param='clean_01', level='A')
        self.assertEqual([], CoreCommons.list_formatter(dpm.get_intent(level='A').keys()))
        self.assertEqual(['0'], CoreCommons.list_formatter(dpm.get_intent(level='B').keys()))

    def test_get_intent(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        dpm.set_intent({'remove_head': {'headers': ['one', 'two']}, 'to_str': {'headers': ['five', 'six']}})
        dpm.set_intent({'clean_03': {'headers': 'h3'}}, level='B')
        result = dpm.get_intent(level='A', order=0, intent='remove_head')
        self.assertDictEqual({'remove_head': {'headers': ['one', 'two']}}, result)
        result = dpm.get_intent(level='A', order=1, intent='to_str')
        self.assertDictEqual({'to_str' : {'headers': ['five', 'six']}}, result)
        result = dpm.get_intent(level='A')
        self.assertEqual(['0', '1'], CoreCommons.list_formatter(result.keys()))
        result = dpm.get_intent(level='B')
        self.assertEqual(['0'], CoreCommons.list_formatter(result.keys()))
        self.assertEqual(['clean_03'], CoreCommons.list_formatter(result.get('0').keys()))
        result = dpm.get_intent()
        self.assertEqual(['A', 'B'], CoreCommons.list_formatter(result.keys()))

    def test_set_intent_description(self):
        pm = ControlPropertyManager('task')
        with self.assertRaises(ValueError) as context:
            pm.set_intent_description(level='control', text='Some intent level text')
        self.assertTrue(f"The intent_level 'control' can not be found " in str(context.exception))
        pm.set_intent({'clean_03': {'headers': 'h3'}}, level='control')
        pm.set_intent_description(level='control', text='Some intent level text')
        self.assertEqual(['Some intent level text'], pm.get_knowledge(catalog='intent', label='control',as_list=True))
        pm.remove_intent(level='control')
        self.assertEqual([], pm.get_knowledge(catalog='intent', label='control', as_list=True))

    def test_intent_report(self):
        dpm = ControlPropertyManager('test_abstract_properties')
        dpm.set_intent({'clean_03': {'headers': 'h3'}}, level='A')
        dpm.set_intent({'clean_02': {'dtype': ['number', 'str'], 'headers': 'h1'}}, level='C')
        dpm.set_intent({'clean_01': {'headers': 'h1'}}, order=-1)
        dpm.set_intent({'clean_01': {'headers': 'h3'}}, order=-1)
        report = dpm.report_intent()
        self.assertEqual(['clean_03', 'clean_01', 'clean_01', 'clean_02'], report['intent'])
        self.assertEqual(['A', 'A', 'A', 'C'], report['level'])
        self.assertEqual(['0', '1', '2', '0'], report['order'])
        self.assertEqual(["headers='h3'"], report['parameters'][0])
        self.assertEqual(["headers='h1'"], report['parameters'][1])
        self.assertEqual(["headers='h3'"], report['parameters'][2])
        self.assertEqual(['default']*4, report['creator'])
        self.assertEqual( ["dtype=['number', 'str']", "headers='h1'"], report['parameters'][3])
        report = dpm.report_intent(level_label='features')
        self.assertEqual(['features', 'order', 'intent', 'parameters', 'creator'], list(report))
        dpm.set_intent_description(level='A', text='Initial cleaning')
        report = dpm.report_intent(as_description=True, level_label='features')
        self.assertEqual(['features', 'description'], list(report))
        self.assertEqual(['A', 'C'], report['features'])
        self.assertEqual(['Initial cleaning', ''], report['description'])

    def test_run_book(self):
        pm = ControlPropertyManager('test_abstract_properties')
        result = pm.report_run_book()
        self.assertDictEqual({'name': [], 'run_book': []}, result)
        pm.set_run_book(book_name='run1', run_levels=[1,5,3])
        self.assertEqual([1,5,3], pm.get_run_book('run1'))
        pm.set_run_book(book_name='run2', run_levels=['A', 'B'])
        self.assertEqual(['A', 'B'], pm.get_run_book('run2'))
        pm.set_run_book(book_name='run1', run_levels=[1, 2, 3])
        self.assertEqual([1,2,3], pm.get_run_book('run1'))
        result = pm.report_run_book()
        self.assertDictEqual({'name': ['run2', 'run1'], 'run_book': [['A', 'B'], [1, 2, 3]]}, result)

    def test_schema(self):
        pm = ControlPropertyManager('test_abstract_properties')
        self.assertEqual([], pm.canonical_schemas)
        self.assertFalse(pm.has_canonical_schema(name='root'))
        schema1 =  dict({'att1': ['type', 'nulls']})
        schema2 =  dict({'A': [1, 0.98]})
        pm.set_canonical_schema('root', schema1)
        pm.set_canonical_schema('schema', schema2)
        self.assertTrue(pm.has_canonical_schema(name='root'))
        self.assertTrue(pm.has_canonical_schema(name='schema'))
        self.assertDictEqual(schema1, pm.get_canonical_schema('root'))
        self.assertDictEqual(schema2, pm.get_canonical_schema('schema'))
        self.assertCountEqual(['root', 'schema'], pm.canonical_schemas)
        self.assertFalse(pm.remove_canonical_schema('name'))
        self.assertTrue(pm.remove_canonical_schema('schema'))
        self.assertCountEqual(['root'], pm.canonical_schemas)
        self.assertDictEqual(schema1, pm.get_canonical_schema('root'))
        self.assertDictEqual({}, pm.get_canonical_schema('schema'))

    def test_file_pattern(self):
        pm = ControlPropertyManager('task')
        result = pm.file_pattern(name='source')
        self.assertEqual('hadron_control_task_source.parquet', result)
        result = pm.file_pattern(name='source', file_type='parquet')
        self.assertEqual('hadron_control_task_source.parquet', result)
        result = pm.file_pattern(name='source', stamped="hours")
        self.assertEqual('hadron_control_task_source${TO_HOURS}.parquet', result)
        result = pm.file_pattern(name='source', versioned=True, stamped="minutes")
        self.assertEqual('hadron_control_task_source${VERSION}${TO_MINUTES}.parquet', result)
        result = pm.file_pattern(name='source', versioned=True)
        self.assertEqual('hadron_control_task_source${VERSION}.parquet', result)
        # Check it works with connector contract
        cc = ConnectorContract(uri=result, module_name="", handler="", version="v1.01")
        self.assertEqual("hadron_control_task_source_v1.01.parquet", cc.uri)
        # test prefix & suffix & suffix_name
        result = pm.file_pattern(name='source', prefix="prefix_")
        self.assertEqual('prefix_source.parquet', result)
        result = pm.file_pattern(name='source')
        self.assertEqual('hadron_control_task_source.parquet', result)
        result = pm.file_pattern(name='source', suffix="?name=fred&type=male")
        self.assertEqual('hadron_control_task_source.parquet?name=fred&type=male', result)
        result = pm.file_pattern(name='source', prefix="prefix_", suffix="?name=fred&type=male")
        self.assertEqual('prefix_source.parquet?name=fred&type=male', result)
        # test path
        result = pm.file_pattern(name='source', path="s3://bucket/path")
        self.assertEqual('s3://bucket/path/hadron_control_task_source.parquet', result)
        result = pm.file_pattern(name='source', path=["s3://bucket", "path"])
        self.assertEqual('s3://bucket/path/hadron_control_task_source.parquet', result)
        # test project
        result = pm.file_pattern(name='source', project='tester')
        self.assertEqual('tester_control_task_source.parquet', result)

    def test_modify_connector_uri(self):
        pm = ControlPropertyManager('task')
        connector_name = 'connector'
        uri = "s3://bucket/path/hadron_control_task_connector.parquet?name=fred&value=45"
        cc = ConnectorContract(uri=uri, module_name='module.package', handler='Handler')
        pm.set_connector_contract(connector_name=connector_name, connector_contract=cc)
        control = "http://git.com/project-hadron/hadron_control_task_connector.parquet?name=fred&value=45"
        result = pm.modify_connector_uri(connector_name=connector_name, old_pattern="s3://bucket/path/", new_pattern="http://git.com/project-hadron/")
        self.assertEqual(control, pm.get_connector_contract(connector_name=connector_name).uri)

    def test_modify_connector_aligned(self):
        pm = ControlPropertyManager('task')
        connector_name = 'connector'
        uri = "s3://bucket/path/hadron_control_task_connector.parquet?name=fred&value=45"
        cc = ConnectorContract(uri=uri, module_name='module.package', handler='Handler')
        template = ConnectorContract(uri="http://git.com/hadron/", module_name='module.parquet', handler='PickleHandler',
                                     sep= ',', encoder= 'latin1')
        pm.set_connector_contract(connector_name=connector_name, connector_contract=cc, aligned=True)
        pm.modify_connector_aligned(connector_name=connector_name, template_contract=template)
        control = "http://git.com/hadron/hadron_control_task_connector.parquet?name=fred&value=45"
        result = pm.get_connector_contract(connector_name=connector_name)
        self.assertEqual('module.parquet', result.raw_module_name)
        self.assertEqual('PickleHandler', result.raw_handler)
        self.assertEqual(control, result.raw_uri)
        self.assertDictEqual( {'encoder': 'latin1', 'sep': ','}, result.kwargs)

    def test_reset_each(self):
        pm = ControlPropertyManager('task')
        pm.reset_canonical_schemas()
        pm.reset_connector_contracts()
        pm.reset_intents()
        pm.reset_knowledge()
        pm.reset_run_books()
        pm.reset_snapshots()
        result = pm.get(pm.KEY.contract_key)
        control = {'cleaners': {}, 'description': '', 'status': 'discovery', 'version': '0.0.1', 'connectors': {}, 'intent': {},
                   'meta': {'class': 'ControlPropertyManager', 'module': pm.__module__.split(".")}, 'knowledge': {'schema': {}}, 'run_book': {}, 'snapshot': {}}
        self.assertDictEqual(control, result)

    def test_report_connectors_align(self):
        pm = ControlPropertyManager('task')
        connector_name = 'connector'
        uri = "https://bucket/path/hadron_control_task_connector.parquet?name=fred&value=45"
        cc = ConnectorContract(uri=uri, module_name='module.package', handler='Handler')
        pm.set_connector_contract(connector_name='source', connector_contract=cc)
        pm.set_connector_contract(connector_name='persist', connector_contract=cc, aligned=True)
        result = pm.report_connectors()
        self.assertEqual([False, True], result.get('aligned'))

    def test_report_task(self):
        pm = ControlPropertyManager('task')
        control = ['cleaners', 'description', 'version', 'status', 'connectors', 'intent', 'run_book', 'knowledge']
        self.assertCountEqual(control, list(pm.report_task().keys()))

    def test_report_intent_param(self):
        pm = ControlPropertyManager('task')
        cleaner1 = {'get_number': {'range_value': '0', 'to_value': 24}}
        pm.set_intent(cleaner1, level='age', order=0)
        result = pm.report_intent_params('age')
        self.assertCountEqual(['order', 'intent', 'parameter', 'value'], list(result.keys()))


if __name__ == '__main__':
    unittest.main()


import unittest
import os
from pathlib import Path
import shutil
from datetime import datetime
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from ds_core.handlers.abstract_handlers import ConnectorContract
from ds_core.handlers.event_handlers import EventSourceHandler, EventPersistHandler, EventManager
from ds_core.properties.property_manager import PropertyManager

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatureBuilderTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'parquet'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_event_manager(self):
        tbl = get_table()
        em = EventManager()
        # set
        em.set('task', tbl)
        self.assertTrue(em.is_event('task'))
        self.assertTrue(EventManager().is_event('task'))
        self.assertEqual(['task'], em.event_names())
        self.assertTrue(tbl.equals(em.get('task')))
        # update
        sub_tbl = tbl.drop_columns(['num', 'cat'])
        em.update('task', sub_tbl)
        self.assertFalse(tbl.equals(em.get('task')))
        self.assertTrue(sub_tbl.equals(em.get('task')))
        # delete
        em.delete('task')
        self.assertFalse(em.is_event('task'))

    def test_connector_contract(self):
        tbl = get_table()
        uri = 'event://task/'
        cc = ConnectorContract(uri, 'module_name', 'handler')
        in_handler = EventPersistHandler(cc)
        in_handler.persist_canonical(tbl)
        out_handler = EventSourceHandler(cc)
        result = out_handler.load_canonical()
        self.assertTrue(tbl.equals(result))

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")

def get_table():
    num = pa.array([1.0, None, 5.0, -0.46421, 3.5, 7.233, -2], pa.float64())
    val = pa.array([1, 2, 3, 4, 5, 6, 7], pa.int64())
    date = pc.strptime(["2023-01-02 04:49:06", "2023-01-02 04:57:12", None, None, "2023-01-02 05:23:50", None, None],
                       format='%Y-%m-%d %H:%M:%S', unit='ns')
    text = pa.array(["Blue", "Green", None, 'Red', 'Orange', 'Yellow', 'Pink'], pa.string())
    binary = pa.array([True, True, None, False, False, True, False], pa.bool_())
    cat = pa.array([None, 'M', 'F', 'M', 'F', 'M', 'M'], pa.string()).dictionary_encode()
    return pa.table([num, val, date, text, binary, cat], names=['num', 'int', 'date', 'text', 'bool', 'cat'])


if __name__ == '__main__':
    unittest.main()



import importlib.util
import os
import re
from datetime import datetime
from urllib.parse import parse_qsl, urlparse, urlunparse, urlencode
from abc import ABC, abstractmethod
from typing import Any

__author__ = 'Darryl Oatridge'


class ConnectorContract(object):
    """ a container class for Connector Contract"""
    raw_uri: str = None
    raw_module_name: str = None
    raw_handler: str = None
    raw_kwargs: dict = None
    raw_version: str = None

    def __init__(self, uri: str,  module_name: str, handler: str, version: str=None, **kwargs):
        """ initialisation of the Connector Contract. Though not required, the URI can be considered as a
        URL of the form 'scheme://netloc/path;parameters?query#fragment' Following the syntax specifications
        in RFC 1808. This allows the use of the helper method `parse_uri` and static method `parse_params`

        :param uri: A Uniform Resource Identifier that unambiguously identifies a particular resource
        :param module_name: the module or package name where the handler can be found
        :param handler: the handler for retrieving the resource
        :param version: a version number to pass to the connector
        :param kwargs: (optional) key word arguments to be passed to the handler.
        """
        if not isinstance(uri, str) or len(uri) == 0:
            raise ValueError("The uri must be a valid string")
        self.raw_uri = uri
        self.raw_module_name = module_name
        self.raw_handler = handler
        self.raw_version = version if isinstance(version, str) else "0.0.1"
        self.raw_kwargs = kwargs if isinstance(kwargs, dict) else {}

    @property
    def uri(self) -> str:
        _parsed = self._format_uri(uri_raw=self.raw_uri, version=self.version)
        return self.parse_environ(_parsed)

    @property
    def module_name(self) -> str:
        return self.parse_environ(self.raw_module_name)

    @property
    def handler(self) -> str:
        return self.parse_environ(self.raw_handler)

    @property
    def kwargs(self) -> dict:
        """copy of the private kwargs dictionary"""
        _kwargs = self.raw_kwargs.copy() if isinstance(self.raw_kwargs, dict) else {}
        return self.parse_environ(_kwargs)

    @property
    def version(self) -> str:
        return self.parse_environ(self.raw_version)

    @property
    def path(self) -> str:
        return urlparse(self.uri).path

    @property
    def address(self) -> str:
        return self.parse_address(uri=self.uri, with_credentials=True, with_port=True)

    @property
    def query(self) -> dict:
        """copy of the private query dictionary"""
        _query = dict(parse_qsl(urlparse(self.uri).query))
        if isinstance(_query, dict):
            return _query.copy()
        return {}

    @property
    def schema(self) -> str:
        return urlparse(self.uri).scheme

    @property
    def netloc(self) -> str:
        return urlparse(self.uri).netloc

    @property
    def username(self) -> str:
        return urlparse(self.uri).username

    @property
    def password(self) -> str:
        return urlparse(self.uri).password

    @property
    def hostname(self) -> str:
        return urlparse(self.uri).hostname

    @property
    def port(self) -> int:
        return urlparse(self.uri).port

    @property
    def fragment(self) -> str:
        return urlparse(self.uri).fragment

    @property
    def params(self) -> list:
        """copy of the private params list"""
        _params = urlparse(self.uri).params
        return _params.split(sep=';').copy() if len(_params) > 0 else list()

    def get_key_value(self, key: str, default: Any=None) -> Any:
        """ returns the value for the key in the kwargs or query dictionaries.
        If the key is found in both then the query value is returned

        :param key: the key to look for
        :param default: a default value to return if not found
        :return: the value of the key or the default value if key is not found
        """
        value = self.raw_kwargs.get(key, default)
        return self.query.get(key, value)

    @staticmethod
    def _format_uri(uri_raw: str, version: str):
        """The URI with any modifications"""
        pattern = r'\${([A-Za-z_0-9\-]+)}'
        uri_tags = re.findall(pattern, uri_raw)
        if len(uri_tags) == 0:
            return uri_raw
        tags_dict = ConnectorContract._uri_tag_dict(version)
        pattern = r'\${([A-Za-z_0-9\-]+)}'
        return re.sub(pattern, lambda m: tags_dict.get(m.group(1), "${" + m.group(1) + "}"), uri_raw)

    @staticmethod
    def _uri_tag_dict(version: str) -> dict:
        """returns a dictionary of URI tags and their substitute"""
        return {'VERSION': f"_{version}",
                'TO_DAYS': datetime.now().strftime("_%Y%m%d"),
                'TO_HOURS': datetime.now().strftime("_%Y%m%d%H"),
                'TO_MINUTES': datetime.now().strftime("_%Y%m%d%H%M"),
                'TO_SECONDS': datetime.now().strftime("_%Y%m%d%H%M%S"),
                'TO_NS': datetime.now().strftime("_%Y%m%d%H%M%S%f")}

    @staticmethod
    def uri_tags() -> list:
        """returns the list of valid uri substitute tags"""
        return list(ConnectorContract._uri_tag_dict(version='').keys())

    @staticmethod
    def parse_environ(parse: [str, dict]):
        """parse a string replacing environment variables in the path"""
        pattern = r'\${([A-Za-z_0-9\-]+)}'
        if isinstance(parse, str):
            for label in re.findall(pattern, parse):
                if label not in os.environ.keys():
                    raise EnvironmentError(f"when parsing the connector attribute '{parse}' The environment "
                                           f"variable '{label}' was referenced but not set in the environment")
            return re.sub(pattern, lambda m: os.getenv(m.group(1), parse), parse)
        if isinstance(parse, dict):
            for key, value in parse.items():
                for label in re.findall(pattern, f"{key}{value}"):
                    if label not in os.environ.keys():
                        raise EnvironmentError(f"when parsing key '{key}', value '{value}' The environment "
                                               f"variable '{label}' was referenced but not set in the environment")

                key = re.sub(pattern, lambda m: os.getenv(m.group(1), key), key) if isinstance(key, str) else key
                value = re.sub(pattern,
                               lambda m: os.getenv(m.group(1), value), value) if isinstance(value, str) else value
                parse.update({key: value})
        return parse

    @staticmethod
    def parse_address_elements(uri: str, with_credentials: bool=None, with_port: bool=None) -> tuple:
        """ utility method to extract the address elements (schema, netloc, path) from a URI.
        Optionally the credentials and/or port can be excluded from the netloc

        :param uri: the URI to parse
        :param with_credentials: (optional) if to include the credentials. Default is True
        :param with_port: (optional) if to include the port. Default is True
        :return: a tuple of (schema, netloc, path)
        """
        _address = ConnectorContract.parse_address(uri=uri, with_credentials=with_credentials, with_port=with_port)
        parse_address = urlparse(_address)
        return tuple([parse_address.scheme, parse_address.netloc, parse_address.path])

    @staticmethod
    def parse_query(uri: str) -> dict:
        """ utility method to extract the query element from a URI

        :param uri: the URI to parse
        :return:
        """
        parse_url = urlparse(uri)
        _query = dict(parse_qsl(parse_url.query))
        return _query

    @staticmethod
    def parse_address(uri: str, with_credentials: bool=None, with_port: bool=None) -> str:
        """ utility method to extract the address from a URI, removing params, query and fragment. optionally
        the credentials and port can be excluded

        :param uri: the URI to parse
        :param with_credentials: (optional) if to include the credentials. Default is True
        :param with_port: (optional) if to include the port. Default is True
        :return: the full address string
        """
        with_credentials = with_credentials if isinstance(with_credentials, bool) else True
        with_port = with_port if isinstance(with_port, bool) else True
        if uri is None or not uri:
            raise ValueError("The uri must be a valid string with a length greater than 0")
        parse_url = urlparse(uri)
        _netloc = parse_url.hostname
        if with_credentials and isinstance(parse_url.username, str) and isinstance(parse_url.password, str):
            _credentials = ":".join([parse_url.username, parse_url.password])
            _netloc = '@'.join([_credentials, _netloc])
        if with_port and isinstance(parse_url.port, int):
            _netloc = ':'.join([_netloc, str(parse_url.port)])
        return urlunparse((parse_url.scheme, _netloc, parse_url.path, '', '', ''))

    @staticmethod
    def unparse_address(scheme: str, netloc: str, path: str, params: str=None, query: str=None, fragment: str=None):
        """ returns reconstructed URI from the parsed elements"""
        params = params if isinstance(params, str) else ''
        query = query if isinstance(query, str) else ''
        fragment = fragment if isinstance(fragment, str) else ''
        return urlunparse((scheme, netloc, path, params, query, fragment))

    @staticmethod
    def encode_query(**kwargs):
        """given kwargs, synthesize a query string"""
        return urlencode(kwargs)

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the Connector Contract"""
        uri_dict = {'address': self.address, 'schema': self.schema, 'netloc': self.netloc, 'path': self.path,
                    'params': self.params, 'query': self.query, 'fragment': self.fragment}
        for attr in ['username', 'password', 'hostname', 'port']:
            attr_value = eval('self.{}'.format(attr))
            if attr_value is not None:
                uri_dict.update({attr: attr_value})
        rtn_dict = {'uri': self.uri, 'uri_parsed': uri_dict, 'module_name': self.module_name,
                    'handler': self.handler, 'version': self.version, 'kwargs': self.kwargs}
        return rtn_dict

    def to_raw_dict(self) -> dict:
        """Returns a dictionary representation of the Connector Contract"""
        rtn_dict = {'raw_uri': self.raw_uri, 'raw_module_name': self.raw_module_name, 'raw_handler': self.raw_handler,
                    'raw_version': self.raw_version, 'raw_kwargs': self.raw_kwargs}
        return rtn_dict

    def __len__(self):
        return self.to_raw_dict().__len__()

    def __str__(self):
        return self.to_dict().__str__()

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.to_raw_dict().__str__()}"

    def __eq__(self, other):
        return self.to_raw_dict().__eq__(other.to_raw_dict())

    def __setattr__(self, key, value):
        if self.to_raw_dict().get(key, None) is None:
            super().__setattr__(key, value)
        else:
            raise AttributeError("The attribute '{}' is immutable once set and can not be changed".format(key))

    def __delattr__(self, item):
        raise AttributeError("{} is an immutable class and attributes can't be removed".format(self.__class__.__name__))


class AbstractSourceHandler(ABC):

    def __init__(self, connector_contract: ConnectorContract):
        self._contract = connector_contract

    @property
    def connector_contract(self) -> ConnectorContract:
        return self._contract

    def set_connector_contract(self, source_contract: ConnectorContract):
        self._contract = source_contract

    @abstractmethod
    def supported_types(self) -> list:
        pass

    @abstractmethod
    def exists(self) -> bool:
        pass

    @abstractmethod
    def has_changed(self) -> bool:
        pass

    @abstractmethod
    def reset_changed(self, changed: bool=False):
        pass

    @abstractmethod
    def load_canonical(self, **kwargs) -> Any:
        pass


class AbstractPersistHandler(AbstractSourceHandler):

    @abstractmethod
    def persist_canonical(self, canonical: Any, **kwargs) -> bool:
        pass

    @abstractmethod
    def remove_canonical(self, **kwargs) -> bool:
        pass

    @abstractmethod
    def backup_canonical(self, canonical: Any, uri: str, **kwargs) -> bool:
        """ creates a backup of the canonical to an alternative uri

        :param canonical: the canonical to back up
        :param uri: an alternative uri to the one in the ConnectorContract
        :param kwargs: if given, these kwargs are used as a replacement of the connector kwargs
        :return: True if successful
        """
        pass


class HandlerFactory(object):

    @staticmethod
    def check_module(module_name: str) -> bool:
        module_spec = importlib.util.find_spec(module_name)
        if module_spec is None:
            return False
        return True

    @staticmethod
    def check_handler(module_name: str, handler: str):
        try:
            module_spec = importlib.util.find_spec(module_name)
        except ModuleNotFoundError:
            return False
        if module_spec is None:
            return False
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        if handler not in dir(module):
            return False
        return True

    @staticmethod
    def get_module(module_name: str):
        module_spec = importlib.util.find_spec(module_name)
        if module_spec is None:
            raise ModuleNotFoundError(f"The module '{module_name}' could not be found")
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        return module

    @staticmethod
    def instantiate(connector_contract: ConnectorContract) -> [AbstractSourceHandler, AbstractPersistHandler]:
        module_name = connector_contract.module_name
        handler = connector_contract.handler

        # check module
        module_spec = importlib.util.find_spec(module_name)
        if module_spec is None:
            raise ModuleNotFoundError(f"The module '{module_name}' could not be found")

        # check handler
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        if handler not in dir(module):
            raise ImportError(f"The handler '{handler}' could not be found in the module '{module_name}'")

        # create instance of handler
        local_kwargs = locals().get('kwargs') if 'kwargs' in locals() else dict()
        local_kwargs['module'] = module
        local_kwargs['connector_contract'] = connector_contract
        instance = eval(f'module.{handler}(connector_contract)', globals(), local_kwargs)
        if not isinstance(instance, (AbstractSourceHandler, AbstractPersistHandler)):
            raise TypeError(f"The handler '{handler}' in package {module_name} could not be instanciated as a handler")
        return instance



import os
import platform
from abc import ABC, abstractmethod
from typing import Any

from ds_core.components.core_commons import CoreCommons

from ds_core.intent.abstract_intent import AbstractIntentModel
from ds_core.properties.abstract_properties import AbstractPropertyManager
from ds_core.handlers.abstract_handlers import ConnectorContract, HandlerFactory
from ds_core.handlers.abstract_handlers import AbstractPersistHandler, AbstractSourceHandler

__author__ = 'Darryl Oatridge'


class AbstractComponent(ABC):
    """ Abstract AI Single Task Application Component (AI-STAC) component class provides all the basic building blocks
    of a components build including property management, augmented knowledge notes and parameterised intent pipeline.

    For convenience there are two Factory Initialisation methods available `from_env(...)` and `from_uri(...)` the
    second being an abstract method. This factory method initialises the concrete PropertyManager and IntentModel
    classes and should use the parent `_init_properties(...)` methods to set the properties connector.

    As an example concrete implementation of these methods:

    .. code-block:: py3

        @classmethod
        def from_uri(cls, task_name: str, uri_pm_path: str, username: str, uri_pm_repo: str=None,
                     pm_file_type: str=None, pm_module: str=None, pm_handler: str=None, pm_kwargs: dict=None,
                     default_save=None, reset_templates: bool=None, template_path: str=None, template_module: str=None,
                     template_source_handler: str=None, template_persist_handler: str=None, align_connectors: bool=None,
                     default_save_intent: bool=None, default_intent_level: bool=None, order_next_available: bool=None,
                     default_replace_intent: bool=None, has_contract: bool=None):
            pm_file_type = pm_file_type if isinstance(pm_file_type, str) else 'parquet'
            pm_module = pm_module if isinstance(pm_module, str) else cls.DEFAULT_MODULE
            pm_handler = pm_handler if isinstance(pm_handler, str) else cls.DEFAULT_PERSIST_HANDLER
            _pm = ExamplePropertyManager(task_name=task_name, username=username)
            _intent_model = ExampleIntentModel(property_manager=_pm, default_save_intent=default_save_intent,
                                               default_intent_level=default_intent_level,
                                               order_next_available=order_next_available,
                                               default_replace_intent=default_replace_intent)
            super()._init_properties(property_manager=_pm, uri_pm_path=uri_pm_path, default_save=default_save,
                                     uri_pm_repo=uri_pm_repo, pm_file_type=pm_file_type, pm_module=pm_module,
                                     pm_handler=pm_handler, pm_kwargs=pm_kwargs, has_contract=has_contract)
            return cls(property_manager=_pm, intent_model=_intent_model, default_save=default_save,
                       reset_templates=reset_templates, template_path=template_path, template_module=template_module,
                       template_source_handler=template_source_handler, template_persist_handler=template_persist_handler,
                       align_connectors=align_connectors)

    """

    # primary connectors
    CONNECTOR_SOURCE = 'primary_source'
    CONNECTOR_PERSIST = 'primary_persist'
    # report connectors
    REPORT_SCHEMA = 'primary_schema'
    REPORT_INTENT = 'intent'
    REPORT_NOTES = 'notes'
    REPORT_CONNECTORS = 'connectors'
    REPORT_RUNBOOK = 'runbook'
    REPORT_ENVIRON = 'environ'
    # template connectors
    # TEMPLATE_SOURCE = 'template_source'
    # TEMPLATE_PERSIST = 'template_persist'
    # default connectors module and handlers
    DEFAULT_MODULE = 'ds_core.handlers.event_handlers'
    DEFAULT_SOURCE_HANDLER = 'EventSourceHandler'
    DEFAULT_PERSIST_HANDLER = 'EventPersistHandler'

    def __init__(self, property_manager: Any, intent_model: Any, default_save: bool=None, reset_templates: bool=None,
                 template_path: str = None, template_module: str = None, template_source_handler: str = None,
                 template_persist_handler: str = None, align_connectors: bool=None):
        """ initialisation of the abstract components providing both the property manager and the components
        parameterizable intent model. The optional parameters allow default references to be overridden by a
        concrete implementations of the abstract.
        The default module and handlers replace the root default static values for DEFAULT_MODULE,
        DEFAULT_SOURCE_HANDLER, DEFAULT_PERSIST_HANDLER and provide implementation specific default references but are
        also used in methods where the module and handlers are optional parameters.
        The default save allows a component to be run in memory or persisted as a default behaviour.

        :param property_manager: The contract property manager instance for this components
        :param intent_model: the model codebase containing the parameterizable intent
        :param default_save: (optional) The default behaviour of persisting the contracts:
                    - if True: all contract properties are persisted
                    - if False: The connector contracts are kept in memory (useful for restricted file systems)

        :param reset_templates: (optional) reset connector templates from environ variables. Default True
        :param template_path: (optional) a template path to use if the environment variable does not exist
        :param template_module: (optional) a template module to use if the environment variable does not exist
        :param template_source_handler: (optional) a template source handler to use if no environment variable
        :param template_persist_handler: (optional) a template persist handler to use if no environment variable
        :param align_connectors: (optional) resets aligned connectors to the template. default Default True
        """
        if not isinstance(property_manager, AbstractPropertyManager):
            raise ValueError("The contract_pm must be a concrete implementation of the AbstractPropertyManager")
        if not isinstance(intent_model, AbstractIntentModel):
            raise ValueError("The intent_model must be a concrete implementation of the AbstractIntent")
        # set instance references
        self._component_pm = property_manager
        self._intent_model = intent_model
        self._default_save = default_save if isinstance(default_save, bool) else True
        # align templates and connectors
        reset_templates = reset_templates if isinstance(reset_templates, bool) else True
        align_connectors = align_connectors if isinstance(align_connectors, bool) else True
        if reset_templates:
            # setup the templates values
            if os.access('./', os.W_OK | os.X_OK):
                _root = './'
            else:
                _root = os.environ['AppData'] if platform.system().lower().startswith("Windows") else '/tmp'
            _default_path = os.getenv('HADRON_DEFAULT_PATH') \
                if 'HADRON_DEFAULT_PATH' in os.environ else os.path.join(_root, 'hadron', 'data')
            _schema, _, _ = ConnectorContract.parse_address_elements(uri=_default_path)
            _default_module, _default_source, _default_persist = self._from_handler(_schema)
            template_path = template_path if isinstance(template_path, str) else _default_path
            template_module = template_module if isinstance(template_module, str) else _default_module
            template_source_handler = template_source_handler \
                if isinstance(template_source_handler, str) else _default_source
            template_persist_handler = template_persist_handler \
                if isinstance(template_persist_handler, str) else _default_persist
            if align_connectors:
                # remove existing
                if self.pm.has_connector(self.pm.TEMPLATE_SOURCE):
                    self.remove_connector_contract( self.pm.TEMPLATE_SOURCE)
                if self.pm.has_connector(self.pm.TEMPLATE_PERSIST):
                    self.remove_connector_contract( self.pm.TEMPLATE_PERSIST)
                # reset the template Connector Contracts
                scc = ConnectorContract(uri=template_path, module_name=template_module, handler=template_source_handler)
                self.add_connector_contract(connector_name=self.pm.TEMPLATE_SOURCE, connector_contract=scc)
                pcc = ConnectorContract(uri=template_path, module_name=template_module, handler=template_persist_handler)
                self.add_connector_contract(connector_name=self.pm.TEMPLATE_PERSIST, connector_contract=pcc)
                # reset the connector
                self.reset_template_connectors()
        # just make sure everything is persisted
        self.pm_persist()
        return

    @classmethod
    @abstractmethod
    def from_uri(cls, task_name: str, uri_pm_path: str, creator: str, uri_pm_repo: str=None, pm_file_type: str=None,
                 pm_module: str=None, pm_handler: str=None, pm_kwargs: dict=None, default_save=None,
                 reset_templates: bool=None, template_path: str=None, template_module: str=None,
                 template_source_handler: str=None, template_persist_handler: str=None, align_connectors: bool=None,
                 default_save_intent: bool=None, default_intent_level: bool=None, order_next_available: bool=None,
                 default_replace_intent: bool=None, has_contract: bool=None):
        """ Class Factory Method to instantiates the components application. The Factory Method handles the
        instantiation of the Properties Manager, the Intent Model and the persistence of the uploaded properties.
        See class inline docs for an example method

         :param task_name: The reference name that uniquely identifies a task or subset of the property manager
         :param uri_pm_path: A URI that identifies the resource path for the property manager.
         :param creator: A user name for this task activity.
         :param uri_pm_repo: (optional) A repository URI to initially load the property manager but not save to.
         :param pm_file_type: (optional) defines a specific file type for the property manager
         :param pm_module: (optional) the module or package name where the handler can be found
         :param pm_handler: (optional) the handler for retrieving the resource
         :param pm_kwargs: (optional) a dictionary of kwargs to pass to the property manager
         :param default_save: (optional) if the configuration should be persisted. default to 'True'
         :param reset_templates: (optional) reset connector templates from environ variables. Default True
         :param template_path: (optional) a template path to use if the environment variable does not exist
         :param template_module: (optional) a template module to use if the environment variable does not exist
         :param template_source_handler: (optional) a template source handler to use if no environment variable
         :param template_persist_handler: (optional) a template persist handler to use if no environment variable
         :param align_connectors: (optional) resets aligned connectors to the template. default Default True
         :param default_save_intent: (optional) The default action for saving intent in the property manager
         :param default_intent_level: (optional) the default level intent should be saved at
         :param order_next_available: (optional) if the default behaviour for the order should be next available order
         :param default_replace_intent: (optional) the default replace existing intent behaviour
         :param has_contract: (optional) indicates the instance should have a property manager domain contract
         :return: the initialised class instance
         """
        return cls

    @classmethod
    def _init_properties(cls, property_manager: AbstractPropertyManager, uri_pm_path: str, default_save: bool=None,
                         uri_pm_repo: str=None, pm_file_type: str=None, pm_module: str=None, pm_handler: str=None,
                         has_contract: bool=None, pm_kwargs: dict=None):
        """ initialisation and set up of the property connector contract into the property manager instance.
        This should be used as part of the component initialisation

        :param property_manager: the instance of the property manager
        :param uri_pm_path: the URI path to where property contract are help.
        :param default_save: (optional) if the configuration should be persisted. default to 'True'
        :param uri_pm_repo: The read only repo link that points to the raw data path to the contracts repo directory
        :param pm_file_type: (optional) defines a specific file type for the property manager
        :param pm_module: (optional) the module or package name where the handler can be found
        :param pm_handler: (optional) the handler for retrieving the resource
        :param has_contract: (optional) indicates the instance should have a property manager domain contract
        :param pm_kwargs: (optional) a dictionary of kwargs to pass to the property manager
        :return: a tuple of the selected pm_module_name, pm_handler and pm_file_type
        """
        if not isinstance(uri_pm_path, str) or len(uri_pm_path) == 0:
            raise ValueError("The URI must be a valid string representation of a URI")
        _schema, _netloc, _path = ConnectorContract.parse_address_elements(uri=uri_pm_path)
        _file_type = pm_file_type if isinstance(pm_file_type, str) else 'parquet'
        has_contract = has_contract if isinstance(has_contract, bool) else True
        default_save = default_save if isinstance(default_save, bool) else True
        if isinstance(pm_module, str) and isinstance(pm_handler, str):
            _module_name = pm_module
            _handler = pm_handler
        else:
            _module_name, _, _handler = cls._from_handler(schema=_schema)
            _address = ConnectorContract.parse_address(uri=uri_pm_path)
        _path = os.path.join(_path, property_manager.pm_file_pattern(file_type=_file_type))
        _prop_uri = ConnectorContract.unparse_address(scheme=_schema, netloc=_netloc, path=_path)
        _query_kw = ConnectorContract.parse_query(uri=uri_pm_path)
        kwargs = pm_kwargs if isinstance(pm_kwargs, dict) else {}
        kwargs.update(_query_kw)
        _connector = ConnectorContract(uri=_prop_uri, module_name=_module_name, handler=_handler,
                                       version=property_manager.version, **kwargs)
        # make a copy to set as the ongoing connector if repo is used
        _pm_connector = _connector
        property_manager.set_property_connector(connector_contract=_pm_connector)
        if isinstance(uri_pm_repo, str):
            _r_schema, _r_netloc, _r_path = ConnectorContract.parse_address_elements(uri=uri_pm_repo)
            _r_path = os.path.join(_r_path, property_manager.pm_file_pattern(file_type=_file_type))
            _r_prop_uri = ConnectorContract.unparse_address(scheme=_r_schema, netloc=_r_netloc, path=_r_path)
            _r_query_kw = ConnectorContract.parse_query(uri=uri_pm_repo)
            _connector = ConnectorContract(uri=_r_prop_uri, module_name=_module_name, handler=_handler,
                                           version=property_manager.version, **_r_query_kw)
        if HandlerFactory.instantiate(_connector).exists():
            if not has_contract:
                # if set to not have a contract then reset this contract
                property_manager.reset_all()
                property_manager.persist_properties(connector_contract=_pm_connector, only_branch=True)
                # reset to overwrite anything loaded
                property_manager.set_property_connector(connector_contract=_pm_connector)
            else:
                # set to the temporary property manager
                property_manager.set_property_connector(connector_contract=_connector)
                try:
                    property_manager.load_properties()
                except [KeyError, IOError]:
                    raise ConnectionError("Unable to retrieve the persisted properties, file might be corrupted "
                                          "or of a different format")
                # reset to overwrite anything loaded
                property_manager.set_property_connector(connector_contract=_pm_connector)
        elif has_contract and not HandlerFactory.instantiate(_connector).exists():
            raise FileNotFoundError(f"The property manager domain contact '{_connector.uri}' could not be found."
                                    f" If this is a new contract, set 'has_contract' to False")
        if default_save:
            property_manager.persist_properties(connector_contract=_pm_connector)
        return _module_name, _handler, _file_type

    @classmethod
    def from_memory(cls, has_contract: bool=None, default_save_intent: bool=None, default_intent_level: bool=None,
                    order_next_available: bool=None, default_replace_intent: bool=None, **kwargs):
        """ Class Factory Method that creates a light touch in memory instance that leaves no residue when closed.
        This factory method can load a reference contract from a remote repo as a foundation.

         :param default_save_intent: (optional) The default action for saving intent in the property manager
         :param default_intent_level: (optional) the default level intent should be saved at
         :param order_next_available: (optional) if the default behaviour for the order should be next available order
         :param default_replace_intent: (optional) the default replace existing intent behaviour
         :param has_contract: (optional) indicates the instance should have a property manager domain contract
         :param kwargs: to pass to the property ConnectorContract as its kwargs
         :return: the initialised class instance
        """
        # create a unique task name
        has_contract = has_contract if isinstance(has_contract, bool) else False
        default_save_intent = default_save_intent if isinstance(default_save_intent, bool) else False
        return cls.from_env(task_name='temporary_in_memory_instance', default_save=False, has_contract=has_contract,
                            default_save_intent=default_save_intent, default_intent_level=default_intent_level,
                            order_next_available=order_next_available, default_replace_intent=default_replace_intent,
                            **kwargs)

    @classmethod
    def from_env(cls, task_name: str, default_save=None, reset_templates: bool=None, align_connectors: bool=None,
                 default_save_intent: bool=None, default_intent_level: bool=None, order_next_available: bool=None,
                 default_replace_intent: bool=None, uri_pm_repo: str=None, has_contract: bool=None, **kwargs):
        """ Class Factory Method that builds the connector handlers taking the property contract path from
        the ``os.environ['HADRON_PM_PATH']`` or, if not found, uses the system default,
            - for Linux and IOS '/tmp/components/contracts
            - for Windows ``os.environ['AppData']\\components\\contracts``

        The following environment variables can be set:
            - HADRON_PM_PATH: the property contract path, if not found, uses the system default
            - HADRON_PM_REPO: the property contract should be initially loaded from a read only repo site such as github
            - HADRON_PM_TYPE: a file type for the property manager. If not found sets as parquet
            - HADRON_PM_MODULE: a default module package, if not set uses component default
            - HADRON_PM_HANDLER: a default handler. if not set uses component default

        This method calls to the Factory Method ``from_uri(...)`` returning the initialised class instance

        :param task_name: The reference name that uniquely identifies a task or subset of the property manager
        :param default_save: (optional) if the configuration should be persisted
        :param reset_templates: (optional) reset connector templates from environ variables. Default True
        :param align_connectors: (optional) resets aligned connectors to the template. default Default True
        :param default_save_intent: (optional) The default action for saving intent in the property manager
        :param default_intent_level: (optional) the default level intent should be saved at
        :param order_next_available: (optional) if the default behaviour for the order should be next available order
        :param default_replace_intent: (optional) the default replace existing intent behaviour
        :param uri_pm_repo: The read only repo link that points to the raw data path to the contracts repo directory
        :param has_contract: (optional) indicates the instance should have a property manager domain contract
        :param kwargs: to pass to the property ConnectorContract as its kwargs
        :return: the initialised class instance
        """
        pm_file_type = os.environ.get('HADRON_PM_TYPE', 'parquet')
        pm_uri = os.environ.get('HADRON_PM_PATH', None)
        if isinstance(uri_pm_repo, str):
            pm_repo = uri_pm_repo
            default_save = False
        else:
            pm_repo = os.environ.get('HADRON_PM_REPO', None)
        if pm_uri is None:
            if os.access('.', os.W_OK | os.X_OK):
                pm_uri = os.path.join('.', 'hadron', 'contracts')
            elif platform.system().lower().startswith("Windows"):
                pm_uri = os.path.join(os.environ['AppData'], 'hadron', 'contracts')
            else:
                pm_uri = os.path.join('/tmp', 'hadron', 'contracts')
        creator = os.environ.get('HADRON_CREATOR', os.environ.get('USER', 'Unknown'))
        pm_module = os.environ.get('HADRON_PM_MODULE', None)
        pm_handler = os.environ.get('HADRON_PM_HANDLER', None)
        pm_kwargs = kwargs if isinstance(kwargs, dict) and len(kwargs) > 0 else None
        return cls.from_uri(task_name=task_name, uri_pm_path=pm_uri, uri_pm_repo=pm_repo, creator=creator,
                            has_contract=has_contract, pm_file_type=pm_file_type, pm_module=pm_module,
                            pm_handler=pm_handler, pm_kwargs=pm_kwargs, default_save=default_save,
                            reset_templates=reset_templates, align_connectors=align_connectors,
                            default_save_intent=default_save_intent, default_intent_level=default_intent_level,
                            order_next_available=order_next_available, default_replace_intent=default_replace_intent)

    @classmethod
    def _from_handler(cls, schema: str) -> (str, str, str):
        """ Class Factory Method that builds the connector handlers for the schema."""
        schema = schema if isinstance(schema, str) else ""
        for _package in ['ds_core', 'ds_capability']:
            _module_name = f'{_package}.handlers.{schema.lower()}_handlers'
            _source_handler = f'{schema.title()}SourceHandler'
            _persist_handler = f'{schema.title()}PersistHandler'
            if HandlerFactory.check_handler(_module_name, _persist_handler):
                return _module_name, _source_handler, _persist_handler
        return cls.DEFAULT_MODULE, cls.DEFAULT_SOURCE_HANDLER, cls.DEFAULT_PERSIST_HANDLER

    @classmethod
    def scratch_pad(cls):
        """ A class method to use the Components intent methods as a scratch pad"""
        return cls.from_memory().intent_model

    """
        PROPERTY MANAGER SECTION
    """

    def _get_environ(self, level: str, is_source: bool=None, default: str=None):
        """ returns the preference environ value based on type and if source or persist

        :param level: the level of environ, options are 'path', 'module' or  'handler'
        :param is_source: if this is for source or persist, default is persist
        :return: returns the default value if not found
        """
        if level not in ['path', 'module', 'handler']:
            raise ValueError(f"The type '{level}' is not supported, options are 'path', 'module' or 'handler'")
        handle = 'SOURCE' if isinstance(is_source, bool) and is_source else 'PERSIST'
        manager = f'HADRON_{self.pm.manager_name()}'.upper()
        task = f'HADRON_{self.pm.manager_name()}_{self.pm.task_name}'.upper()
        # task and manager
        result = os.environ.get(f"{task}_{handle}_{level}".upper(),
                                os.environ.get(f"{task}_{level}".upper(),
                                               os.environ.get(f"{manager}_{handle}_{level}".upper(),
                                                              os.environ.get(f"{manager}_{level}".upper(), None))))
        # now try the defaults
        if result is None:
            result = os.environ.get(f"HADRON_DEFAULT_{handle}_{level}".upper(),
                                    os.environ.get(f"HADRON_DEFAULT_{level}".upper(), None))
        return default if result is None else result

    def report_environ(self, hide_not_set: bool=True):
        """returns a report on the foundation environment variables"""
        report = dict()
        for task in [f'{self.pm.manager_name()}_{self.pm.task_name}', f'{self.pm.manager_name()}', 'default']:
            for level in ['path', 'module', 'handler']:
                for handle in ['_source', '_persist', '']:
                    environ = f'HADRON_{task}{handle}_{level}'.upper()
                    if os.environ.get(environ, None) is None and hide_not_set:
                        continue
                    report.update({environ: os.environ.get(environ, 'not used')})
        pc = self.pm.get_connector_contract(self.pm.CONNECTOR_PM_CONTRACT)
        report.update({'HADRON_PM_PATH': os.environ.get('HADRON_PM_PATH', pc.path)})
        report.update({'HADRON_PM_REPO': os.environ.get('HADRON_PM_REPO', 'not used')})
        report.update({'HADRON_PM_TYPE': os.environ.get('HADRON_PM_TYPE', 'default')})
        report.update({'HADRON_PM_MODULE': os.environ.get('HADRON_PM_MODULE', 'default')})
        report.update({'HADRON_PM_HANDLER': os.environ.get('HADRON_PM_HANDLER', 'default')})
        report.update({'HADRON_CREATOR': os.environ.get('HADRON_CREATOR', 'account default')})
        return report

    @property
    def intent_model(self):
        """The intent model instance"""
        return self._intent_model

    @property
    def pm(self):
        """The properties manager instance"""
        return self._component_pm

    def pm_persist(self, save=None):
        """Saves the current configuration to file"""
        if not isinstance(save, bool):
            save = self._default_save
        if save and self.pm.has_connector(self.pm.CONNECTOR_PM_CONTRACT):
            self.pm.persist_properties(only_branch=True)
        return

    @property
    def pm_name(self) -> str:
        """The contract name of this transition instance"""
        return self._component_pm.contract_name

    def pm_reset(self, save: bool=None):
        """ resets the contract back to a default. This does not remove the Property Manager Connector Contract or
        any snapshots

        :param save: override of the default save action set at initialisation.
        """
        if not isinstance(save, bool):
            save = self._default_save
        source_template = self.pm.get_connector_contract(connector_name=self.pm.TEMPLATE_SOURCE)
        persist_template = self.pm.get_connector_contract(connector_name=self.pm.TEMPLATE_PERSIST)
        self.pm.reset_all()
        self.add_connector_contract(connector_name=self.pm.TEMPLATE_SOURCE, connector_contract=source_template)
        self.add_connector_contract(connector_name=self.pm.TEMPLATE_PERSIST, connector_contract=persist_template)
        self.pm_persist(save)
        return

    def pm_transfer(self, transfer_connector: [str, ConnectorContract]):
        """ Takes a copy of the pm contract and saves it to a new location defined by the connector contract.
        This can be used to publish a property manager to a new location, change its format or as a backup

        :param transfer_connector: the name of an existing connector contract or a ConnectorContract
        """
        if not isinstance(transfer_connector, (str, ConnectorContract)):
            raise TypeError("The transfer_connector must be either a name of a connector or a ConnectorContract")
        connector_contract = transfer_connector
        if isinstance(transfer_connector, str):
            connector_contract = self.pm.get_connector_contract(transfer_connector)
        self.pm.persist_properties(connector_contract=connector_contract)
        return

    """
        CONNECTOR CONTRACTS SECTION
    """
    def add_connector_uri(self, connector_name: str, uri: str, save: bool=None, template_aligned: bool=None, **kwargs):
        """ Sets the contract giving the full uri path. This is a shortcut of set_source_contract(...), not
        requiring a ConnectorContract to be set up and using the default module and handler values.

        :param connector_name: the name or label to identify and reference the connector
        :param uri: a fully qualified uri of the source data
        :param template_aligned: the connector aligns with the template so changes to the template
        :param save: (optional) if True, save to file. Default is True
        """
        if not isinstance(uri, str) or len(uri) == 0:
            raise ValueError("The URI must be a valid string representation of a URI")
        template_aligned = template_aligned if isinstance(template_aligned, bool) else False
        _schema, _netloc, _path = ConnectorContract.parse_address_elements(uri=uri)
        _module_name, _, _handler = self._from_handler(schema=_schema)
        connector_contract = ConnectorContract(uri=uri, module_name=_module_name, handler=_handler,
                                               version=self.pm.version, **kwargs)
        self.add_connector_contract(connector_name=connector_name, connector_contract=connector_contract,
                                    template_aligned=template_aligned, save=save)
        return self

    def add_connector_contract(self, connector_name: str, connector_contract: ConnectorContract,
                               template_aligned: bool=None, save: bool=None):
        """ Sets a named connector contract

        :param connector_name: the name or label to identify and reference the connector
        :param connector_contract: a Connector Contract for the properties persistence
        :param template_aligned: the connector aligns with the template so changes to the template
        :param save: override of the default save action set at initialisation.
        :return: if load is True, returns a Pandas.DataFrame else None
        """
        if self.pm.has_connector(connector_name):
            self.pm.remove_connector_contract(connector_name)
        self.pm.set_connector_contract(connector_name=connector_name, connector_contract=connector_contract,
                                       aligned=template_aligned)
        self.pm_persist(save)
        return self

    def add_connector_source(self, connector_name: str, uri_file: str, save: bool=None, **kwargs):
        """ Adds a connector using settings from the self.pm.TEMPLATE_SOURCE template connector.

        :param connector_name: the name or label to identify and reference the connector
        :param uri_file: the name of the file to append to the end of the default path
        :param save: override of the default save action set at initialisation.
        :param kwargs: any kwargs to add to the default connector
        :return:
        """
        self.add_connector_from_template(connector_name=connector_name, uri_file=uri_file,
                                         template_name=self.pm.TEMPLATE_SOURCE, save=save, **kwargs)
        return self

    def add_connector_persist(self, connector_name: str, uri_file: str, save: bool=None, **kwargs):
        """ Adds a connector using settings from the self.pm.TEMPLATE_PERSIST template connector.
        self.pm.TEMPLATE_PERSIST are added at initialisation

        :param connector_name: the name or label to identify and reference the connector
        :param uri_file: the name of the file to append to the end of the default path
        :param save: override of the default save action set at initialisation.
        :param kwargs: any kwargs to add to the default connector
        :return:
        """
        self.add_connector_from_template(connector_name=connector_name, uri_file=uri_file,
                                         template_name=self.pm.TEMPLATE_PERSIST, save=save, **kwargs)
        return self

    def add_connector_from_template(self, connector_name: str, uri_file: str, template_name: str,  save: bool=None,
                                    **kwargs):
        """ Adds a connector using settings from a template connector. By default a self.TEMPLATE_SOURCE and
        self.TEMPLATE_PERSIST are added at initialisation

        :param connector_name: the name or label to identify and reference the connector
        :param uri_file: the name of the file to append to the end of the default path
        :param save: override of the default save action set at initialisation.
        :param kwargs: any kwargs to add to the default connector
        :return:
        """
        if not self.pm.has_connector(connector_name=template_name):
            raise ValueError(f"The template connector '{template_name}' could not be found")
        template = self.pm.get_connector_contract(connector_name=template_name)
        uri = os.path.join(template.raw_uri, uri_file)
        if not isinstance(kwargs, dict):
            kwargs = {}
        template.raw_kwargs.update(kwargs)
        cc = ConnectorContract(uri=uri, module_name=template.raw_module_name, handler=template.raw_handler,
                               version=self.pm.version, **kwargs)
        self.add_connector_contract(connector_name=connector_name, connector_contract=cc, template_aligned=True,
                                    save=save)
        return self

    def reset_template_connectors(self, save: bool=None):
        """ resets connector contracts with template path and handler where they are template aligned.
        (see `set_connector_aligned`)

        :param save: override of the default save action set at initialisation.
        """
        connector_names =  self.pm.connector_contract_list
        for name in connector_names:
            if name in [self.pm.TEMPLATE_SOURCE, self.pm.TEMPLATE_PERSIST]:
                continue
            if self.pm.has_connector(name):
                if self.pm.is_connector_contract_aligned(name):
                    instance = HandlerFactory.instantiate(self.pm.get_connector_contract(connector_name=name))
                    if isinstance(instance, AbstractPersistHandler):
                        if self.pm.has_connector(self.pm.TEMPLATE_PERSIST):
                            persist_template = self.pm.get_connector_contract(self.pm.TEMPLATE_PERSIST)
                            self.pm.modify_connector_aligned(connector_name=name, template_contract=persist_template)
                    elif isinstance(instance, AbstractSourceHandler):
                        if self.pm.has_connector(self.pm.TEMPLATE_SOURCE):
                            source_template = self.pm.get_connector_contract(self.pm.TEMPLATE_SOURCE)
                            self.pm.modify_connector_aligned(connector_name=name, template_contract=source_template)
        self.pm_persist(save)
        return

    def remove_connector_contract(self, connector_name: str, save: bool=None):
        """removes a named connector contract

        :param connector_name: the name or label to identify and reference the connector
        :param save: override of the default save action set at initialisation.
        """
        if not isinstance(save, bool):
            save = self._default_save
        self.pm.remove_connector_contract(connector_name)
        self.pm_persist(save)
        return

    def set_connector_version(self, connector_names: [str, list], version: str, save: bool=None):
        """ modifies the uri of a connector contract and resets

        :param connector_names: a name or list of names of connector contract to modify
        :param version: the new version number
        :param save: override of the default save action set at initialisation.
        """
        if not isinstance(save, bool):
            save = self._default_save
        for name in self.pm.list_formatter(connector_names):
            self.pm.set_connector_version(connector_name=name, version=version)
        self.pm_persist(save)
        return

    def set_connector_aligned(self, connector_names: [str, list], aligned: bool, save: bool=None):
        """ modifies the uri of a connector contract and resets

        :param connector_names: a name or list of names of connector contract to modify
        :param aligned: if the connector contract is aligned to the template connector contract
        :param save: override of the default save action set at initialisation.
        """
        if not isinstance(save, bool):
            save = self._default_save
        for name in self.pm.list_formatter(connector_names):
            self.pm.set_connector_aligned(connector_name=name, aligned=aligned)
        self.pm_persist(save)
        return self

    def set_source(self, uri_file: str, save: bool=None, **kwargs):
        """sets the source contract CONNECTOR_SOURCE using the TEMPLATE_SOURCE connector contract,

        :param uri_file: the uri_file is appended to the template path
        :param save: (optional) if True, save to file. Default is True
        """
        self.add_connector_from_template(connector_name=self.CONNECTOR_SOURCE, uri_file=uri_file,
                                         template_name=self.pm.TEMPLATE_SOURCE, save=save, **kwargs)
        return self

    def set_source_uri(self, uri: str, save: bool=None, template_aligned: bool=None, **kwargs):
        """ Sets the source contract giving the full uri path. This is a shortcut of set_source_contract(...), not
        requiring a ConnectorContract to be set up and using the default module and handler values.

        :param uri: a fully qualified uri of the source data
        :param template_aligned: the connector aligns with the template so changes to the template
        :param save: (optional) if True, save to file. Default is True
        """
        template_aligned = template_aligned if isinstance(template_aligned, bool) else False
        self.add_connector_uri(connector_name=self.CONNECTOR_SOURCE, save=save, uri=uri,
                               template_aligned=template_aligned, **kwargs)
        return self

    def set_source_contract(self, connector_contract: ConnectorContract, template_aligned: bool=None, save: bool=None):
        """ Sets the source contract using the class CONNECTOR_SOURCE constant

        :param connector_contract: a Connector Contract for the source data
        :param template_aligned: the connector aligns with the template so changes to the template
        :param save: (optional) if True, save to file. Default is True
        """
        template_aligned = template_aligned if isinstance(template_aligned, bool) else False
        self.add_connector_contract(connector_name=self.CONNECTOR_SOURCE, connector_contract=connector_contract,
                                    template_aligned=template_aligned, save=save)
        return self

    def get_persist_contract(self) -> ConnectorContract:
        """ gets the persist connector contract that can be used as the next chain source. If
            the uri contains environment variables it is NOT parsed at load
        """
        return self.pm.get_connector_contract(self.CONNECTOR_PERSIST)

    def get_persist_uri(self) -> str:
        """ gets the persist connector contract uri that be used as the next chain source. If
            the uri contains environment variables it is parsed
        """
        return self.pm.get_connector_contract(self.CONNECTOR_PERSIST).raw_uri

    def set_persist(self, uri_file: str=None, save: bool=None, **kwargs):
        """sets the persist contract CONNECTOR_PERSIST using the TEMPLATE_PERSIST connector contract

        :param uri_file: (optional) the uri_file is appended to the template path
        :param save: (optional) if True, save to file. Default is True
        """
        file_pattern = self.pm.file_pattern(name=self.CONNECTOR_PERSIST)
        uri_file = uri_file if isinstance(uri_file, str) else file_pattern
        self.add_connector_from_template(connector_name=self.CONNECTOR_PERSIST, uri_file=uri_file,
                                         template_name=self.pm.TEMPLATE_PERSIST, save=save, **kwargs)
        return self

    def set_persist_uri(self, uri: str, save: bool=None, template_aligned: bool=None, **kwargs):
        """ Sets the persist contract giving the full uri path. This is a shortcut of set_persist_contract(...), not
        requiring a ConnectorContract to be set up and using the default module and handler values.

        :param uri: a fully qualified uri of the persist data
        :param template_aligned: the connector aligns with the template so changes to the template
        :param save: (optional) if True, save to file. Default is True
        """
        template_aligned = template_aligned if isinstance(template_aligned, bool) else False
        self.add_connector_uri(connector_name=self.CONNECTOR_PERSIST, save=save, uri=uri,
                               template_aligned=template_aligned, **kwargs)
        return self

    def set_persist_contract(self, connector_contract: ConnectorContract, save: bool=None):
        """ Sets the persist contract.

        :param connector_contract: a Connector Contract for the persisted data
        :param save: (optional) if True, save to file. Default is True
        """
        self.add_connector_contract(connector_name=self.CONNECTOR_PERSIST, connector_contract=connector_contract,
                                    save=save)
        return self

    def set_report_persist(self, reports: [str, list], project: str=None, path: [str, list]=None, prefix: str=None,
                           suffix: str=None, file_type: str=None, versioned: bool=None, stamped: str=None,
                           save: bool=None, **kwargs) -> list:
        """sets the report persist using the TEMPLATE_PERSIST connector contract, there are preset constants that
        should be used. These constance can be in the form <class>.REPORT_<NAME> or <instance>.REPORT_<NAME>
        where <NAME> is the name of the report and can be found in this class. Examples of reports might be:

        .. code-block:: py3

            Transition.REPORT_SCHEMA
            [self.REPORT_NOTES, self.REPORT_SCHEMA]
            [builder.REPORT_NOTES, {'report': builder.REPORT_SCHEMA, 'uri_file': '<file_name>'}]
            [{'report': Wrangle.REPORT_NOTES, 'file_type': 'parquet'}]
            [{'report': self.REPORT_SCHEMA, 'file_type': 'csv', 'versioned': True, 'stamped': 'days'}]

        if a report is presented as dict, the method signature parameters will be overwritten by the report values.
        This allows globally parameter to apply generally but allow single reports to be modified at a granular level.

        to ensure dict reports have the correct keys the util method 'report2dict(...)' can be used.

        :param reports: (optional) the name(s) of the report connector to set (see class REPORT_* constants)
        :param project: (optional) an alternative project string that replaces 'hadron'
        :param path: (optional) a file path that precedes the prefix and file pattern. uses os.path.join so takes a list
        :param prefix: (optional) a prefix to put at the front of the file pattern to replace the default
        :param suffix: (optional) a suffix to put at the end of the file pattern and extension
        :param file_type: (optional) a global file extension to the default 'parquet' format
        :param versioned: (optional) if all reports should include a version
        :param stamped: (optional) A string of the timestamp options ['days', 'hours', 'minutes', 'seconds', 'ns']
        :param save: (optional) if True, save to file. Default is True
        :param kwargs: (optional) additional parameters to send as kwargs for the Connect Contract
        :return: a list of connector names created from the reports
        """
        if not isinstance(reports, (str, list)):
            raise TypeError(f"The report names must be either a name reference, or a list of str and/or dict. "
                            f"'{type(reports)}' type was passed")
        rtn_reports = []
        for _report in self.pm.list_formatter(reports):
            if not isinstance(_report, (str, dict)):
                raise TypeError(f"The report type {type(_report)} is an unsupported type. Must be string or dict")
            if isinstance(_report, str):
                _report = {'report': _report}
            if not _report.get('report', None):
                raise ValueError(f"The report {_report} must have a 'report' key representing the ref report name")
            name = _report.get('report')
            # if the contact exists then replace it
            if self.pm.has_connector(connector_name=name):
                self.remove_connector_contract(connector_name=name, save=save)
            prefix = _report.get('prefix', prefix)
            suffix = _report.get('suffix', suffix)
            file_type = _report.get('file_type', file_type if isinstance(file_type, str) else 'parquet')
            versioned = _report.get('versioned', versioned if isinstance(versioned, str) else True)
            stamped = _report.get('stamped', stamped if isinstance(stamped, str) else False)
            file_pattern = self.pm.file_pattern(name=name, project=project, path=path, prefix=prefix, suffix=suffix,
                                                file_type=file_type, versioned=versioned, stamped=stamped)
            uri_file = _report.get('uri_file', file_pattern)
            self.add_connector_from_template(connector_name=name, uri_file=uri_file,
                                             template_name=self.pm.TEMPLATE_PERSIST, save=save, **kwargs)
            rtn_reports.append(name)
        return rtn_reports

    """
        INTENT SECTION
    """
    def add_run_book(self, run_levels: [str, list], book_name: str=None, save: bool=None):
        """ sets a named run book, the run levels are a list of levels and the order they are run in

        :param run_levels: the name or list of levels to be run
        :param book_name: (optional) the name of the run_book. defaults to 'primary_run_book'
        :param save: (optional) override of the default save action set at initialisation.
       """
        book_name = book_name if isinstance(book_name, str) else self.pm.PRIMARY_RUN_BOOK
        self.pm.set_run_book(book_name=book_name, run_levels=run_levels)
        self.pm_persist(save)

    def add_run_book_level(self, run_level: str, book_name: str=None, save: bool=None):
        """ adds a single runlevel to the end of a run_book. If the name already exists it will be replaced

        :param run_level: the run_level to add.
        :param book_name: (optional) the name of the run_book. defaults to 'primary_run_book'
        :param save: (optional) override of the default save action set at initialisation.
       """
        book_name = book_name if isinstance(book_name, str) else self.pm.PRIMARY_RUN_BOOK
        if self.pm.has_run_book(book_name=book_name):
            run_levels = self.pm.get_run_book(book_name=book_name)
            while run_level in run_levels:
                run_levels.remove(run_level)
            run_levels.append(run_level)
        else:
            run_levels = run_level
        self.pm.set_run_book(book_name=book_name, run_levels=run_levels)
        self.pm_persist(save)

    def remove_run_book(self, book_name: str=None, save: bool=None) -> bool:
        """ removes named run book. If no runbook is given then all run books are removed

        :param book_name: (optional) the name of the run_book. defaults to primary_run_book'
        :param save: (optional) override of the default save action set at initialisation.
        :return: True if removed, False if not
        """
        if isinstance(book_name, str):
            result = self.pm.remove_run_book(book_name=book_name)
        else:
            run_books = self.pm.get(self.pm.KEY.run_book_key, {})
            if isinstance(run_books, dict):
                for book in tuple(run_books.keys()):
                    self.pm.remove_run_book(book_name=book)
            result = True
        self.pm_persist(save)
        return result

    def remove_intent(self, intent_param: [str, dict]=None, level: [int, str]=None, save: bool=None):
        """ removes part or all the intent contract.
            - If no params all intent is removed
            - if only intent then all references in all params of that named intent will be removed
            - if only level then that level is removed
            - if both level and intent then that specific intent on that level is removed

        :param intent_param: (optional) removes the method contract
        :param level: (optional) removes the level contract
        :param save: (optional) override of the default save action set at initialisation.
        :return: True if removed, False if not
        """
        if isinstance(intent_param, (str, dict)) or isinstance(level, (int, str)):
            result = self.pm.remove_intent(intent_param=intent_param, level=level)
        else:
            for intent in tuple(self.pm.get_intent().keys()):
                self.pm.remove_intent(level=intent)
            result = True
        self.pm_persist(save)
        return result

    def add_intent_level_description(self, level: [int, str], text: str, save: bool=None):
        """ sets description to the augmented knowledge 'intent' to a level

        :param level: the intent level to add the comment to
        :param text: the description text
        :param save: (optional) override of the default save action set at initialisation.
        """
        self.pm.set_intent_description(level=level, text=text)
        self.pm_persist(save)
        return

    """
        CANONICAL SECTION
    """
    def load_canonical(self, connector_name: str, reset_changed: bool=None, has_changed: bool=None,
                       return_empty: bool=None, **kwargs) -> Any:
        """returns the canonical of the referenced connector

        :param connector_name: the name or label to identify and reference the connector
        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        if self.pm.has_connector(connector_name):
            has_changed = has_changed if isinstance(has_changed, bool) else False
            reset_changed = reset_changed if isinstance(reset_changed, bool) else False
            handler = self.pm.get_connector_handler(connector_name)
            if has_changed and handler.exists() == False:
                if isinstance(return_empty, bool) and return_empty:
                    return {}
                raise ConnectionAbortedError("The connector name {} has been aborted as the canonical to load "
                                             "does not exist".format(connector_name))
            if has_changed and handler.has_changed() == False:
                if isinstance(return_empty, bool) and return_empty:
                    return {}
                raise ConnectionAbortedError("The connector name {} has been aborted as the canonical to load "
                                             "has not changed".format(connector_name))
            canonical = handler.load_canonical(**kwargs)
            handler.reset_changed(changed=reset_changed)
            return canonical
        raise ConnectionError("The connector name {} can't be found.".format(connector_name))

    def persist_canonical(self, connector_name: str, canonical: Any, **kwargs):
        """persists the canonical to the referenced connector. same as save_canonical

        :param connector_name: the name or label to identify and reference the connector
        :param canonical: the canonical data to persist
        :param kwargs: arguments to be passed to the handler on persist
        """
        if self.pm.has_connector(connector_name):
            handler = self.pm.get_connector_handler(connector_name)
            handler.persist_canonical(canonical, **kwargs)
            return
        raise ConnectionError("The connector name {} can't be found.".format(connector_name))

    def save_canonical(self, connector_name: str, canonical: Any, **kwargs):
        """saves the canonical to the referenced connector. Same as persist_canonical

        :param connector_name: the name or label to identify and reference the connector
        :param canonical: the canonical data to persist
        :param kwargs: arguments to be passed to the handler on persist
        """
        self.persist_canonical(connector_name=connector_name, canonical=canonical, **kwargs)
        return

    def backup_canonical(self, connector_name: str, canonical: Any, uri: str, **kwargs):
        """persists the canonical to the referenced connector as a backup using the URI to
        replace the current Connector Contract URI.

        :param connector_name: the name or label to identify and reference the connector
        :param canonical: the canonical data to persist
        :param uri: an alternative uri to the one in the ConnectorContract
        :param kwargs: arguments to be passed to the handler on persist
        """
        if self.pm.has_connector(connector_name):
            _handler = self.pm.get_connector_handler(connector_name)
            _cc = self.pm.get_connector_contract(connector_name)
            _address = _cc.parse_address(uri=_cc.uri)
            _path, _, _ext = _address.rpartition('.')
            _handler.backup_canonical(canonical=canonical, uri=uri, **kwargs)
            return
        raise ConnectionError("The connector name {} was not found.".format(connector_name))

    def remove_canonical(self, connector_name: str, **kwargs):
        """removes the current persisted canonical.

        :param connector_name: the name or label to identify and reference the connector
        :param kwargs: arguments to be passed to the handler on remove
        """
        if self.pm.has_connector(connector_name):
            handler = self.pm.get_connector_handler(connector_name)
            handler.remove_canonical(**kwargs)
            return
        raise ConnectionError("The connector name {} was not found.".format(connector_name))

    """
        COMPONENT INFO SECTION FOR STATUS, DESCRIPTION AND VERSIONS
    """
    def set_version(self, version: str, save=None):
        """ sets the version
        :param version: the version to be set
        :param save: override of the default save action set at initialisation.
        """
        if not isinstance(save, bool):
            save = self._default_save
        self.pm.set_version(version=version)
        self.pm_persist(save)
        return

    def set_status(self, status: str, save=None):
        """ sets the status of this component task. Suggested status might be 'discovery', 'stable', 'production'
        :param status: the status to be set,
        :param save: override of the default save action set at initialisation.
        """
        if not isinstance(save, bool):
            save = self._default_save
        self.pm.set_status(status=status)
        self.pm_persist(save)
        return

    def set_description(self, description: str, save=None):
        """ sets the description of this component task
        :param description: a brief description of this component task
        :param save: override of the default save action set at initialisation.
        """
        if not isinstance(save, bool):
            save = self._default_save
        self.pm.set_description(description=description)
        self.pm_persist(save)
        return

    """
        SNAPSHOT SECTION
    """
    def create_snapshot(self, suffix: str=None, version: str=None, save: bool=None):
        """ creates a snapshot of contracts configuration. The name format will be <contract_name>_#<suffix>.

        :param suffix: (optional) adds the suffix to the end of the contract name. if None then date & time used
        :param version: (optional) changes the version number of the current contract
        :param save: override of the default save action set at initialisation.
        :return: a list of current contract snapshots
        """
        if not isinstance(save, bool):
            save = self._default_save
        result = self.pm.set_snapshot(suffix)
        if version is not None:
            self.set_version(version=version)
        self.pm_persist(save)
        return result

    def recover_snapshot(self, snapshot_name: str, overwrite: bool=None, save: bool=None) -> bool:
        """ recovers a snapshot back to the current. The snapshot must be from this root contract.
        by default the original root contract will be overwritten unless the overwrite is set to False.
        if overwrite is False a timestamped snapshot is created

        :param snapshot_name:the name of the snapshot (use self.contract_snapshots to get list of names)
        :param overwrite: (optional) if the original contract should be overwritten. Default to True
        :param save: override of the default save action set at initialisation.
        :return: True if the contract was recovered, else False
        """
        if not isinstance(save, bool):
            save = self._default_save
        result = self.pm.recover_snapshot(snapshot_name=snapshot_name, overwrite=overwrite)
        self.pm_persist(save)
        return result

    def delete_snapshot(self, snapshot_name: str, save: bool=None):
        """ deletes a snapshot

        :param snapshot_name: the name of the snapshot
        :param save: override of the default save action set at initialisation.
        :return: True if successful, False is not found or not deleted
        """
        if not isinstance(save, bool):
            save = self._default_save
        self.pm.remove_snapshot(snapshot_name=snapshot_name)
        self.pm_persist(save)
        return

    """
        NOTES SECTION
    """
    @property
    def notes_catalog(self) -> list:
        """returns the list of allowed catalog names"""
        return self.pm.knowledge_catalog

    def add_notes(self, catalog: str, label: [str, list], text: str, constraints: list=None,
                  save=None):
        """ adds a note to the augmented knowledge.
                if no label is given then a journal date of 'year-month' is provided
                if no catalog is given then the default catalogue name is given

        :param catalog: a catalog name
        :param label: a sub key label or list of labels to separate different information strands
        :param text: the text to add
        :param constraints: (optional) a list of allowed label values, if None then any value allowed
        :param save: if True, save to file. Default is True
        """
        if not isinstance(save, bool):
            save = self._default_save
        self.pm.set_knowledge(catalog=catalog, label=label, text=text, constraints=constraints)
        self.pm_persist(save)

    def remove_notes(self, catalog: str, label: str=None, save=None):
        """ removes a all entries for a labeled note

        :param catalog: the type of note to delete, if left empty all notes removed
        :param label: (Optional) the name of the label to be removed
        :param save: (Optional) if True, save to file. Default is True
        :return: True is successful, False if not
        """
        if not isinstance(save, bool):
            save = self._default_save
        self.pm.remove_knowledge(catalog=catalog, label=label)
        self.pm_persist(save)

    def upload_notes(self, canonical: dict, catalog: str, label_key: str, text_key: str, constraints: list=None,
                     save=None):
        """ Allows bulk upload of notes.

        :param canonical: a dictionary of where the key is the label and value is the text
        :param catalog: (optional) the section these notes should be put in
        :param label_key: the dictionary key name for the labels
        :param text_key: the dictionary key name for the text
        :param constraints: (optional) the limited list of acceptable labels. If not in list then ignored
        :param save: if True, save to file. Default is True
        """
        if label_key not in canonical.keys():
            raise ValueError(f"The label_key '{label_key}' is not a key of the canonical")
        if text_key not in canonical.keys():
            raise ValueError(f"The text_key '{text_key}' is not a key of the canonical")
        self.pm.bulk_upload_knowledge(canonical, catalog=catalog, label_key=label_key, text_key=text_key,
                                      constraints=constraints)
        self.pm_persist(save)

    @staticmethod
    def report2dict(report: str, file_type: str=None, versioned: bool=None, stamped: str=None, prefix: str=None,
                    suffix: str=None, path: [str, list]=None) -> dict:
        """ a utility method to help build analytics conditions by aligning method parameters with dictionary format.

        :param report: The name of the report
        :param file_type: (optional) an alternative file extension to the default 'parquet' format
        :param versioned: (optional) if the component version should be included as part of the pattern
        :param stamped: (optional) A string of the timestamp options ['days', 'hours', 'minutes', 'seconds', 'ns']
        :param prefix: (optional) a prefix to put at the front of the file pattern to replace the default
        :param suffix: (optional) a suffix to put at the end of the file pattern and extension
        :param path: (optional) a file path that precedes the prefix and file pattern. uses os.path.join so takes a list
        :return: a dictionary for an individual element
        """
        return CoreCommons.param2dict(**locals())
